from time import time
from re import findall
from typing import Union

from . import UserProfile

class PrepareMessage:
    def __init__(self, **kwargs) -> None:
        self.base_message = {
            "content": kwargs.get("content", None),
            "mediaType": kwargs.get("mediaType", 0),
            "type": kwargs.get("type", 0),
            "clientRefId": int(time() / 10 % 1000000000),
            "timestamp": int(time() * 1000)
            }
        [self.base_message.update({key: kwargs[key]}) for key in kwargs]

    def json(self):
        """`JSON` - returns the raw data."""
        return self.base_message


class MessageAuthor:
    def __init__(self, data: Union[dict, str]) -> None:
        self.data = data

    @property
    def uid(self) -> Union[str, None]:
        """
        `uid` - returns the user id of the author.
            - `str` - The user id of the author.
            - `None` - If the data is `None`.

        """
        return self.data.get("uid")
    
    @property
    def userId(self) -> Union[str, None]:
        """
        `userId` - returns the user id of the author.
            - `str` - The user id of the author.
            - `None` - If the data is `None`.

        """
        return self.uid
    
    @property
    def status(self) -> Union[int, None]:
        """
        `status` - returns the status of the author.
            - `int` - The status of the author.
            - `None` - If the data is `None`.

        """
        return self.data.get("status")
    
    @property
    def icon(self) -> Union[str, None]:
        """
        `icon` - returns the icon of the author.
            - `str` - The icon of the author.
            - `None` - If the data is `None`.

        """
        return self.data.get("icon")
    
    @property
    def avatar(self) -> Union[str, None]:
        """
        `avatar` - returns the icon of the author.
            - `str` - The icon of the author.
            - `None` - If the data is `None`.

        """
        return self.icon
    
    @property
    def reputation(self) -> Union[int, None]:
        """
        `reputation` - returns the reputation of the author.
            - `int` - The reputation of the author.
            - `None` - If the data is `None`.

        """
        return self.data.get("reputation")
    
    @property
    def role(self) -> Union[int, None]:
        """
        `role` - returns the role of the author.
            - `int` - The role of the author.
            - `None` - If the data is `None`.

        """
        return self.data.get("role")
    
    @property
    def nickname(self) -> Union[str, None]:
        """
        `nickname` - returns the nickname of the author.
            - `str` - The nickname of the author.
            - `None` - If the data is `None`.

        """
        return self.data.get("nickname")
    
    @property
    def username(self) -> Union[str, None]:
        """
        `username` - returns the nickname of the author.
            - `str` - The nickname of the author.
            - `None` - If the data is `None`.

        """
        return self.nickname
    
    @property
    def level(self) -> Union[int, None]:
        """
        `level` - returns the level of the author.
            - `int` - The level of the author.
            - `None` - If the data is `None`.

        """
        return self.data.get("level")
    
    @property
    def accountMembershipStatus(self) -> Union[int, None]:
        """
        `accountMembershipStatus` - returns the account membership status of the author.
            - `int` - The account membership status of the author.
            - `None` - If the data is `None`.

        """
        return self.data.get("accountMembershipStatus")
    
    @property
    def avatarFrame(self) -> Union[str, None]:
        """
        `avatarFrame` - returns the avatar frame of the author.
            - `str` - The avatar frame of the author.
            - `None` - If the data is `None`.

        """
        return self.data.get("avatarFrame")
        
    def json(self) -> Union[dict, str]:
        """`JSON` - returns the raw data."""
        return self.data


class CMessage:
    def __init__(self, data: Union[dict, str]) -> None:
        self.data: dict = data.get('result') or data.get('message', data)

    def return_none(func):
        def wrapper(*args, **kwargs):
            return None if args[0].data is None else func(*args, **kwargs)
        return wrapper
    
    @property
    @return_none
    def includedInSummary(self) -> Union[bool, None]:
        """
        `includedInSummary` - returns if the message is included in the summary.
            - `bool` - If the message is included in the summary.
            - `None` - If the data is `None`.

        """
        return self.data.get("includedInSummary")
    
    @property
    @return_none
    def uid(self) -> Union[str, None]:
        """
        `uid` - returns the user id of the author.
            - `str` - The user id of the author.
            - `None` - If the data is `None`.

        """
        return self.data.get("uid")
    
    @property
    @return_none
    def userId(self) -> Union[str, None]:
        """
        `userId` - returns the user id of the author.
            - `str` - The user id of the author.
            - `None` - If the data is `None`.

        """
        return self.uid
    
    @property
    @return_none
    def author(self) -> Union[dict, None]:
        """
        `author` - returns the author object of the message.
            - `dict` - The author of the message.
            - `None` - If the data is `None`.

        """
        return self.data.get("author")
    
    @property
    @return_none
    def isHidden(self) -> Union[bool, None]:
        """
        `isHidden` - returns if the message is hidden.
            - `bool` - If the message is hidden.
            - `None` - If the data is `None`.

        """
        return self.data.get("isHidden")
    
    @property
    @return_none
    def messageId(self) -> Union[str, None]:
        """
        `messageId` - returns the message id.
            - `str` - The message id.
            - `None` - If the data is `None`.

        """
        return self.data.get("messageId")
    
    @property
    @return_none
    def mediaType(self) -> Union[int, None]:
        """
        `mediaType` - returns the media type of the message.
            - `int` - The media type of the message.
            - `None` - If the data is `None`.

        """
        return self.data.get("mediaType")
    
    @property
    @return_none
    def content(self) -> Union[str, None]:
        """
        `content` - returns the content of the message.
            - `str` - The content of the message.
            - `None` - If the data is `None`.

        """
        return self.data.get("content")
    
    @property
    @return_none
    def clientRefId(self) -> Union[str, None]:
        """
        `clientRefId` - returns the client reference id of the message.
            - `str` - The client reference id of the message.
            - `None` - If the data is `None`.

        """
        return self.data.get("clientRefId")
    
    @property
    @return_none
    def threadId(self) -> Union[str, None]:
        """
        `threadId` - returns the thread id of the message.
            - `str` - The thread id of the message.
            - `None` - If the data is `None`.

        """
        return self.data.get("threadId")
    
    @property
    @return_none
    def chatId(self) -> Union[str, None]:
        """
        `chatId` - returns the chat id of the message.
            - `str` - The chat id of the message.
            - `None` - If the data is `None`.

        """
        return self.threadId
    
    @property
    @return_none
    def createdTime(self) -> Union[str, None]:
        """
        `createdTime` - returns the created time of the message.
            - `str` - The created time of the message.
            - `None` - If the data is `None`.

        """
        return self.data.get("createdTime")
    
    @property
    @return_none
    def extensions(self) -> Union[dict, None]:
        """
        `extensions` - returns the extensions of the message.
            - `dict` - The extensions of the message.
            - `None` - If the data is `None`.

        """
        return self.data.get("extensions")
    
    @property
    @return_none
    def type(self) -> Union[int, None]:
        """
        `type` - returns the type of the message.
            - `int` - The type of the message.
            - `None` - If the data is `None`.

        """
        return self.data.get("type")
    
    @property
    @return_none
    def mediaValue(self) -> Union[str, None]:
        """
        `mediaValue` - returns the media value of the message.
            - `str` - The media value of the message.
            - `None` - If the data is `None`.

        """
        return self.data.get("mediaValue")

    def json(self) -> Union[dict, str]:
        """`JSON` - returns the raw data."""
        return self.data
    
class ReplyMessage:
    def __init__(self, data: Union[dict, str]) -> None:
        self.data = data

    def return_none(func):
        def wrapper(*args, **kwargs):
            return None if args[0].data is None else func(*args, **kwargs)
        return wrapper
    
    @property
    @return_none
    def includedInSummary(self) -> Union[bool, None]:
        """
        `Included in summary` - is a boolean value that determines whether the message is included in the summary of the chat.

        - `True` means that the message is included in the summary of the chat.
        - `False` means that the message is not included in the summary of the chat.
        - `None` means that there was an error while trying to get the value.

        """
        return self.data.get('includedInSummary')
    
    @property
    @return_none
    def uid(self) -> Union[str, None]:
        """
        `UID` - is the user id of the user who sent the message that was replied to.

         THIS IS NOT THE BOT'S USER ID.
            - `str` means that the user id was successfully retrieved.
            - `None` means that there was an error while trying to get the user id.

        """
        return self.data.get('uid')
    
    @property
    @return_none
    def userId(self) -> Union[str, None]:
        """
        `User ID` - is the user id of the user who sent the message that was replied to.

         THIS IS NOT THE BOT'S USER ID.
            - `str` means that the user id was successfully retrieved.
            - `None` means that there was an error while trying to get the user id.

        """
        return self.uid
    
    @property
    @return_none
    def author(self) -> Union[UserProfile, None]:
        """
        `Author` - is the user profile of the user who sent the message that was replied to.

         THIS IS NOT THE BOT'S USER PROFILE.
            - `UserProfile` means that the user profile was successfully retrieved.
            - `None` means that there was an error while trying to get the user profile.

        """
        try:
            return UserProfile(self.data.get('author'))
        except Exception:
            return None
    
    @property
    @return_none
    def isHidden(self) -> Union[bool, None]:
        """
        `Is Hidden` - is a boolean value that determines whether the message that was replied to is hidden.

        - `True` means that the message is hidden.
        - `False` means that the message is not hidden.
        - `None` means that there was an error while trying to get the value.

        """
        return self.data.get('isHidden')
    
    @property
    @return_none
    def messageId(self) -> Union[str, None]:
        """
        `Message ID` - is the message id of the message that was replied to.

        THIS IS NOT THE BOT'S MESSAGE ID.
            - `str` means that the message id was successfully retrieved.
            - `None` means that there was an error while trying to get the message id.

        """
        return self.data.get('messageId')
    
    @property
    @return_none
    def mediaType(self) -> Union[int, None]:
        """
        `Media Type` - is the media type of the message that was replied to.
        #TODO: Add media types
        """
        return self.data.get('mediaType')
    
    @property
    @return_none
    def content(self) -> Union[str, None]:
        """
        `Content` - is the content of the message that was replied to.

        THIS IS NOT THE BOT'S MESSAGE CONTENT.
            - `str` means that the message content was successfully retrieved.
            - `None` means that there was an error while trying to get the message content.

        """
        return self.data.get('content')
    
    @property
    @return_none
    def clientRefId(self) -> Union[int, None]:
        """
        `Client Ref ID` - is the client reference id of the message that was replied to.

        THIS IS NOT THE BOT'S CLIENT REFERENCE ID.
            - `int` means that the client reference id was successfully retrieved.
            - `None` means that there was an error while trying to get the client reference id.

        """
        return self.data.get('clientRefId')
    
    @property
    @return_none
    def threadId(self) -> Union[str, None]:
        """
        `Thread ID` - is the thread / chat id that the message that was replied to was sent in.

            - `str` means that the thread id was successfully retrieved.
            - `None` means that there was an error while trying to get the thread id.

        """
        return self.data.get('threadId')
    
    @property
    @return_none
    def chatId(self) -> Union[str, None]:
        """
        `Chat ID` - is the thread / chat id that the message that was replied to was sent in.

            - `str` means that the thread id was successfully retrieved.
            - `None` means that there was an error while trying to get the thread id.

        """
        return self.threadId
    
    @property
    @return_none
    def createdTime(self) -> Union[str, None]:
        """
        `Created Time` - is the time that the message that was replied to was sent.

            - `str` means that the time was successfully retrieved.
            - `None` means that there was an error while trying to get the time.

        """
        return self.data.get('createdTime')
    
    @property
    @return_none
    def type(self) -> Union[str, None]:
        """
        `Type` - is the type of the message that was replied to.

            - `str` means that the type was successfully retrieved.
            - `None` means that there was an error while trying to get the type.

        """
        return self.data.get('type')
    
    @property
    @return_none
    def mediaValue(self) -> Union[str, None]:
        """
        `Media Value` - is the media value of the message that was replied to.

            - `str` means that the media value was successfully retrieved.
            - `None` means that there was an error while trying to get the media value.

        """
        return self.data.get('mediaValue')
    
    def json(self) -> Union[dict, str]:
        """`JSON` - returns the raw data."""
        return self.data

class CMessageExtensions:
    def __init__(self, data: Union[dict, str]) -> None:
        self.data = data

    @property
    def messageId(self) -> Union[str, None]:
        """
        `Message ID` - is the message id of the message that was replied to.
        
        THIS IS NOT THE BOT'S MESSAGE ID.
            - `str` means that the message id was successfully retrieved.
            - `None` means that there was an error while trying to get the message id.
            
        """
        return self.data.get('replyMessageId')
        
    @property
    def message(self) -> ReplyMessage:
        """
        `Message` - is the message that was replied to.
        
        THIS IS NOT THE BOT'S MESSAGE.
            - `ReplyMessage` means that the message was successfully retrieved.
            - `None` means that there was an error while trying to get the message.
        
        """
        return ReplyMessage(self.data.get('replyMessage'))
    
    def json(self) -> Union[dict, str]:
        """`JSON` - returns the raw data."""
        return self.data

class CMessageAuthorList:
    def __init__(self, data: list) -> None:
        self.data = data

    def return_none(func):
        def wrapper(*args, **kwargs):
            return None if args[0].data is None else func(*args, **kwargs)
        return wrapper
    
    @property
    @return_none
    def status(self) -> Union[str, None]:
        """
        `status` - Returns a list of the status of the users in the chat.
            - `str` means that the status was successfully retrieved.
            - `None` means that there was an error while trying to get the status.

        """
        return [i.get('status') for i in self.data]
    
    @property
    @return_none
    def icon(self) -> Union[str, None]:
        """
        `icon` - Returns a list of the icon of the users in the chat.
            - `str` means that the icon was successfully retrieved.
            - `None` means that there was an error while trying to get the icon.

        """
        return [i.get('icon') for i in self.data]
    
    @property
    @return_none
    def avatar(self) -> Union[str, None]:
        """
        `avatar` - Returns a list of the avatar of the users in the chat.
            - `str` means that the avatar was successfully retrieved.
            - `None` means that there was an error while trying to get the avatar.

        """
        return self.icon
    
    @property
    @return_none
    def reputation(self) -> Union[str, None]:
        """
        `reputation` - Returns a list of the reputation of the users in the chat.
            - `str` means that the reputation was successfully retrieved.
            - `None` means that there was an error while trying to get the reputation.

        """
        return [i.get('reputation') for i in self.data]
    
    @property
    @return_none
    def role(self) -> Union[str, None]:
        """
        `role` - Returns a list of the role of the users in the chat.
            - `str` means that the role was successfully retrieved.
            - `None` means that there was an error while trying to get the role.

        """
        return [i.get('role') for i in self.data]
    
    @property
    @return_none
    def nickname(self) -> Union[str, None]:
        """
        `nickname` - Returns a list of the nickname of the users in the chat.
            - `str` means that the nickname was successfully retrieved.
            - `None` means that there was an error while trying to get the nickname.

        """
        return [i.get('nickname') for i in self.data]
    
    @property
    @return_none
    def username(self) -> Union[str, None]:
        """
        `username` - Returns a list of the username of the users in the chat.
            - `str` means that the username was successfully retrieved.
            - `None` means that there was an error while trying to get the username.

        """
        return self.nickname
    
    @property
    @return_none
    def level(self) -> Union[str, None]:
        """
        `level` - Returns a list of the level of the users in the chat.
            - `str` means that the level was successfully retrieved.
            - `None` means that there was an error while trying to get the level.

        """
        return [i.get('level') for i in self.data]
    
    @property
    @return_none
    def accountMembershipStatus(self) -> Union[str, None]:
        """
        `accountMembershipStatus` - Returns a list of the account membership status of the users in the chat.
            - `str` means that the account membership status was successfully retrieved.
            - `None` means that there was an error while trying to get the account membership status.

        """
        return [i.get('accountMembershipStatus') for i in self.data]
    
    @property
    @return_none
    def avatarFrame(self) -> Union[str, None]:
        """
        `avatarFrame` - Returns a list of the avatar frame of the users in the chat.
            - `str` means that the avatar frame was successfully retrieved.
            - `None` means that there was an error while trying to get the avatar frame.

        """
        return [i.get('avatarFrame') for i in self.data]

    def json(self) -> Union[dict, str]:
        """`JSON` - returns the raw data."""
        return self.data

class CMessages:
    def __init__(self, data: Union[dict, str]) -> None:
        self.data = data.get('result') or data.get('messageList', data)

    def return_none(func):
        def wrapper(*args, **kwargs):
            return None if args[0].data is None else func(*args, **kwargs)
        return wrapper
    
    @property
    @return_none
    def includedInSummary(self) -> Union[bool, None]:
        """
        `includedInSummary` - Returns a list of the included in summary of the messages in the chat.
            - `bool` means that the included in summary was successfully retrieved.
            - `None` means that there was an error while trying to get the included in summary.

        """
        return [i.get('includedInSummary') for i in self.data]
    
    @property
    @return_none
    def uid(self) -> Union[str, None]:
        """
        `uid` - Returns a list of the uid of the messages in the chat.
            - `str` means that the uid was successfully retrieved.
            - `None` means that there was an error while trying to get the uid.

        """
        return [i.get('uid') for i in self.data]
    
    @property
    @return_none
    def userId(self) -> Union[str, None]:
        """
        `userId` - Returns a list of the user id of the messages in the chat.
            - `str` means that the user id was successfully retrieved.
            - `None` means that there was an error while trying to get the user id.

        """
        return self.uid
    
    @property
    @return_none
    def author(self) -> Union[CMessageAuthorList, None]:
        """
        `author` - Returns a list of the author of the messages in the chat.
            - `CMessageAuthorList` means that the author was successfully retrieved.
            - `None` means that there was an error while trying to get the author.

        """
        return CMessageAuthorList([i.get('author') for i in self.data])
    
    @property
    @return_none
    def isHidden(self) -> Union[bool, None]:
        """
        `isHidden` - Returns a list of the is hidden of the messages in the chat.
            - `bool` means that the is hidden was successfully retrieved.
            - `None` means that there was an error while trying to get the is hidden.

        """
        return [i.get('isHidden') for i in self.data]
    
    @property
    @return_none
    def messageId(self) -> Union[str, None]:
        """
        `messageId` - Returns a list of the message id of the messages in the chat.
            - `str` means that the message id was successfully retrieved.
            - `None` means that there was an error while trying to get the message id.

        """
        return [i.get('messageId') for i in self.data]
    
    @property
    @return_none
    def mediaType(self) -> Union[str, None]:
        """
        `mediaType` - Returns a list of the media type of the messages in the chat.
            - `str` means that the media type was successfully retrieved.
            - `None` means that there was an error while trying to get the media type.

        """
        return [i.get('mediaType') for i in self.data]
    
    @property
    @return_none
    def content(self) -> Union[str, None]:
        """
        `content` - Returns a list of the content of the messages in the chat.
            - `str` means that the content was successfully retrieved.
            - `None` means that there was an error while trying to get the content.

        """
        return [i.get('content') for i in self.data]
    
    @property
    @return_none
    def clientRefId(self) -> Union[str, None]:
        """
        `clientRefId` - Returns a list of the client ref id of the messages in the chat.
            - `str` means that the client ref id was successfully retrieved.
            - `None` means that there was an error while trying to get the client ref id.

        """
        return [i.get('clientRefId') for i in self.data]
    
    @property
    @return_none
    def threadId(self) -> Union[str, None]:
        """
        `threadId` - Returns a list of the thread id of the messages in the chat.
            - `str` means that the thread id was successfully retrieved.
            - `None` means that there was an error while trying to get the thread id.

        """
        return [i.get('threadId') for i in self.data]
    
    @property
    @return_none
    def chatId(self) -> Union[str, None]:
        """
        `chatId` - Returns a list of the chat id of the messages in the chat.
            - `str` means that the chat id was successfully retrieved.
            - `None` means that there was an error while trying to get the chat id.

        """
        return self.threadId
    
    @property
    @return_none
    def createdTime(self) -> Union[int, None]:
        """
        `createdTime` - Returns a list of the created time of the messages in the chat.
            - `int` means that the created time was successfully retrieved.
            - `None` means that there was an error while trying to get the created time.

        """
        return [i.get('createdTime') for i in self.data]
    
    @property
    @return_none
    def extensions(self) -> Union[dict, None]:
        """
        `extensions` - Returns a list of the extensions of the messages in the chat.
            - `dict` means that the extensions was successfully retrieved.
            - `None` means that there was an error while trying to get the extensions.

        """
        return [i.get('extensions') for i in self.data]
    
    @property
    @return_none
    def type(self) -> Union[str, None]:
        """
        `type` - Returns a list of the type of the messages in the chat.
            - `str` means that the type was successfully retrieved.
            - `None` means that there was an error while trying to get the type.

        """
        return [i.get('type') for i in self.data]
    
    @property
    @return_none
    def mediaValue(self) -> Union[str, None]:
        """
        `mediaValue` - Returns a list of the media value of the messages in the chat.
            - `str` means that the media value was successfully retrieved.
            - `None` means that there was an error while trying to get the media value.

        """
        return [i.get('mediaValue') for i in self.data]
        
    def json(self) -> Union[dict, str]:
        """`JSON` - returns the raw data."""
        return self.data

class Message:
    def __init__(self, data: Union[dict, str]) -> None:
        self.data: dict = data.get("o", data)

    @property
    def ndcId(self) -> Union[int, None]:
        """
        `ndcId` - is the community id that the message was sent in.
        
            - `int` means that the ndcId was successfully retrieved.
            - `None` means that there was an error while trying to get the ndcId.
        
        """
        return self.data.get('ndcId')
    
    @property
    def comId(self) -> Union[int, None]:
        """
        `comId` - is the community id that the message was sent in.
        
            - `int` means that the comId was successfully retrieved.
            - `None` means that there was an error while trying to get the comId.
            
        """
        return self.ndcId
    
    @property
    def chatMessage(self) -> Union[dict, None]:
        """
        `chatMessage` - is the message that was sent.

            Contains all the information about the message.
            - `dict` means that the chatMessage was successfully retrieved.
            - `None` means that there was an error while trying to get the chatMessage.

        """
        return self.data.get('chatMessage')
    
    @property
    def author(self) -> Union[MessageAuthor, None]:
        """
        `author` - is the author of the message.
        
            - `MessageAuthor` means that the author was successfully retrieved.
            - `None` means that there was an error while trying to get the author.
            
        """
        return MessageAuthor(self.chatMessage.get('author')) if self.chatMessage else None
    
    @property
    def mediaValue(self) -> Union[str, None]:
        """
        `mediaValue` - is the media value of the message.
        
            - `str` means that the mediaValue was successfully retrieved.
            - `None` means that there was an error while trying to get the mediaValue.
            
        """
        return self.chatMessage.get('mediaValue')
    
    @property
    def threadId(self) -> Union[str, None]:
        """
        `threadId` - is the thread id the message was sent in.
        
            - `str` means that the threadId was successfully retrieved.
            - `None` means that there was an error while trying to get the threadId.
            
        """
        return self.chatMessage.get('threadId')
    
    @property
    def chatId(self) -> Union[str, None]:
        """
        `chatId` - is the thread id the message was sent in.
        
            - `str` means that the chatId was successfully retrieved.
            - `None` means that there was an error while trying to get the chatId.
        
        """
        return self.threadId
    
    @property
    def mediaType(self) -> Union[int, None]:
        """
        `mediaType` - is the media type of the message.
        
            - `int` means that the mediaType was successfully retrieved.
            - `None` means that there was an error while trying to get the mediaType.
            
        """
        return self.chatMessage.get('mediaType')
    
    @property
    def content(self) -> Union[str, None]:
        """
        `content` - is the content of the message.
        
            - `str` means that the content was successfully retrieved.
            - `None` means that there was an error while trying to get the content.
        
        """
        return self.chatMessage.get('content')
    
    @property
    def clientRefId(self) -> Union[int, None]:
        """
        `clientRefId` - is the client reference id of the message.
        
            - `int` means that the clientRefId was successfully retrieved.
            - `None` means that there was an error while trying to get the clientRefId.
        
        """
        return self.chatMessage.get('clientRefId')
    
    @property
    def messageId(self) -> Union[str, None]:
        """
        `messageId` - is the message id of the message.
        
            - `str` means that the messageId was successfully retrieved.
            - `None` means that there was an error while trying to get the messageId.
            
        """
        return self.chatMessage.get('messageId')
    
    @property
    def uid(self) -> Union[str, None]:
        """
        `uid` - is the user id of the user who sent the message.
        
            - `str` means that the uid was successfully retrieved.
            - `None` means that there was an error while trying to get the uid.
            
        """
        return self.chatMessage.get('uid')
    
    @property
    def userId(self) -> Union[str, None]:
        """
        `userId` - is the user id of the user who sent the message.
        
            - `str` means that the userId was successfully retrieved.
            - `None` means that there was an error while trying to get the userId.
        
        """
        return self.uid
    
    @property
    def createdTime(self) -> Union[str, None]:
        """
        `createdTime` - is the time the message was sent.
        
            - `str` means that the createdTime was successfully retrieved.
            - `None` means that there was an error while trying to get the createdTime.
        
        """
        return self.chatMessage.get('createdTime')
    
    @property
    def type(self) -> Union[int, None]:
        """
        `type` - is the type of the message.
        
            - `int` means that the type was successfully retrieved.
            - `None` means that there was an error while trying to get the type.
        
        """
        return self.chatMessage.get('type')
    
    @property
    def isHidden(self) -> Union[bool, None]:
        """
        `isHidden` - is whether the message is hidden or not.

            - `bool` means that the isHidden was successfully retrieved.
            - `None` means that there was an error while trying to get the isHidden.

        """
        return self.chatMessage.get('isHidden')
    
    @property
    def includedInSummary(self) -> Union[bool, None]:
        """
        `includedInSummary` - is whether the message is included in the summary or not.
        
            - `bool` means that the includedInSummary was successfully retrieved.
            - `None` means that there was an error while trying to get the includedInSummary.
            
        """
        return self.chatMessage.get('includedInSummary')
    
    @property
    def chatBubbleId(self) -> Union[str, None]:
        """
        `chatBubbleId` - is the chat bubble id of the message.
        
            - `str` means that the chatBubbleId was successfully retrieved.
            - `None` means that there was an error while trying to get the chatBubbleId.
            
        """
        return self.chatMessage.get('chatBubbleId')
    
    @property
    def chatBubbleVersion(self) -> Union[int, None]:
        """
        `chatBubbleVersion` - is the chat bubble version of the message.
        
            - `int` means that the chatBubbleVersion was successfully retrieved.
            - `None` means that there was an error while trying to get the chatBubbleVersion.
            
        """
        return self.chatMessage.get('chatBubbleVersion')

    @property
    def replied(self) -> CMessageExtensions:
        """
        `replied` - Returns the replied message if there was one.
        
            - `CMessageExtensions` means there was a reply.
            - `None` means there was no reply.
            
        """
        return CMessageExtensions(self.chatMessage.get('extensions')) if self.chatMessage.get('extensions') != {} else None
    
    @property
    def extensions(self) -> Union[dict, None]:
        """
        `extensions` - Returns the extensions of the message.
        
            - `dict` means that the extensions was successfully retrieved.
            - `None` means that there was an error while trying to get the extensions.
            
        """
        return self.chatMessage.get('extensions')
    
    @property
    def mentioned_user_ids(self) -> Union[list, None]:
        """
        `mentioned_user_ids` - Returns a list of the mentioned user ids.
        
            - `list` means there were mentioned user ids.
            - `None` means there were no mentioned user ids.
            
        """
        return [
            i.get('uid', None) for i in self.extensions.get('mentionedArray')
            ] if self.extensions.get('mentionedArray') is not None else None
    
    @property
    def mentioned_user_names(self) -> Union[list, None]:
        """
        `mentioned_user_names` - Returns a list of the mentioned user names.
        
            - `list` means there were mentioned user names.
            - `None` means there were no mentioned user names.
            
        """
        return findall(r'@([^\u202c\u202d]+)', self.content) if self.content is not None else None
    
    @property
    def mentioned_dictionary(self) -> Union[dict, None]:
        """
        `mentioned_dictionary` - Returns a dictionary of the mentioned {userid: username}.

            - `dict` means there was a mentioned dictionary.
            - `None` means there was no mentioned dictionary.
        
        """
        return dict(zip(
            self.mentioned_user_ids, self.mentioned_user_names
            )) if self.mentioned_user_ids is not None and self.mentioned_user_names is not None else None

    @property
    def alertOption(self) -> Union[int, None]:
        """
        `alertOption` - is the alert option of the message.
        
            - `int` means that the alertOption was successfully retrieved.
            - `None` means that there was an error while trying to get the alertOption.
            
        """
        return self.chatMessage.get('alertOption')
    
    @property
    def membershipStatus(self) -> Union[int, None]:
        """
        `membershipStatus` - is the membership status of the message.
        
            - `int` means that the membershipStatus was successfully retrieved.
            - `None` means that there was an error while trying to get the membershipStatus.
            
        """
        return self.chatMessage.get('membershipStatus')
        
    def json(self) -> Union[dict, str]:
        """`JSON` - returns the raw data."""
        return self.data

class Channel:
    def __init__(self, data: Union[dict, str]) -> None:
        self.data = data.get("o", data)

    def return_none(func):
        def wrapper(*args, **kwargs):
            return None if args[0].data is None else func(*args, **kwargs)
        return wrapper
    
    @property
    @return_none
    def id(self) -> Union[str, None]:
        """
        `id` - is the id of the channel.

            - `str` means that the id was successfully retrieved.
            - `None` means that there was an error while trying to get the id.

        """
        return self.data.get('id')
    
    @property
    @return_none
    def channelName(self) -> Union[str, None]:
        """
        `channelName` - is the channel name of the channel.

            - `str` means that the channelName was successfully retrieved.
            - `None` means that there was an error while trying to get the channelName.

        """
        return self.data.get('channelName')
    
    @property
    @return_none
    def channelKey(self) -> Union[str, None]:
        """
        `channelKey` - is the channel key of the channel.

            - `str` means that the channelKey was successfully retrieved.
            - `None` means that there was an error while trying to get the channelKey.

        """
        return self.data.get('channelKey')
    
    @property
    @return_none
    def channelUid(self) -> Union[int, None]:
        """
        `channelUid` - is the channel uid of the channel.

            - `int` means that the channelUid was successfully retrieved.
            - `None` means that there was an error while trying to get the channelUid.

        """
        return self.data.get('channelUid')
    
    @property
    @return_none
    def expiredTime(self) -> Union[int, None]:
        """
        `expiredTime` - is the expired time of the channel.

            - `int` means that the expiredTime was successfully retrieved.
            - `None` means that there was an error while trying to get the expiredTime.

        """
        return self.data.get('expiredTime')
    
    @property
    @return_none
    def ndcId(self) -> Union[int, None]:
        """
        `ndcId` - is the community id that the channel is in.

            - `int` means that the ndcId was successfully retrieved.
            - `None` means that there was an error while trying to get the ndcId.

        """
        return self.data.get('ndcId')
    
    @property
    @return_none
    def comId(self) -> Union[int, None]:
        """
        `comId` - is the community id that the channel is in.

            - `int` means that the comId was successfully retrieved.
            - `None` means that there was an error while trying to get the comId.

        """
        return self.ndcId
    
    @property
    @return_none
    def threadId(self) -> Union[str, None]:
        """
        `threadId` - is the thread id of the channel.

            - `str` means that the threadId was successfully retrieved.
            - `None` means that there was an error while trying to get the threadId.

        """
        return self.data.get('threadId')
    
    @property
    @return_none
    def chatId(self) -> Union[str, None]:
        """
        `chatId` - is the chat id of the channel.

            - `str` means that the chatId was successfully retrieved.
            - `None` means that there was an error while trying to get the chatId.

        """
        return self.threadId
    
    def json(self) -> Union[dict, str]:
        """`JSON` - returns the raw data."""
        return self.data





from .enums import *
from .general import *
from .messages import *
from .chat_threads import *
from .userprofile import *
from .wsevents import *
from .handlers import *
from .exceptions import *
from .comments import *
from .api_response import *
from .link_info import *
from .notification import *
from .acm import *
from .member import *
from .bubble import *
from .sticker import *
from .admin_log import *


from .entities import *
from .utilities.generate import *
from typing import Any, Callable, TypeVar


F = TypeVar("F", bound=Callable[..., Any])

class Global:
    def __init__(self) -> None:
        pass


    def authenticated(func: F) -> F:
        """
        A decorator that checks if the client is authenticated.
        
        :param func: The function to decorate.
        :type func: F
        :return: The decorated function.
        :rtype: F
        
        This decorator is used to check if the client is authenticated.
        If the client is not authenticated, a `LoginRequired` exception
        will be raised.
        """
        def wrapper(*args, **kwargs) -> Any:
            try:
                if not args[0].is_authenticated:
                    raise LoginRequired
                return func(*args, **kwargs)
            except AttributeError:
                raise LoginRequired
        return wrapper


    def make_request(self, method: str, url: str, data: dict = None, is_login_required: bool = True) -> dict:
        """
        Makes a request to the API.
        
        :param method: The HTTP method to use.
        :type method: str
        :param url: The URL to make the request to.
        :type url: str
        :param data: The data to send with the request.
        :type data: dict
        :return: The response from the request.
        :rtype: dict
        :param is_login_required: Whether the request requires the client to be logged in.
        :type is_login_required: bool
        
        This method is used to make requests to the API.
        
        Example usage:
        
        >>> client.make_request(
            ...     method = "GET",
            ...     url = "/g/s/user-profile/0000-000000-000000-0000-000000"
            ...     is_login_required = True
            ... )
            
        This will make a GET request to the URL `/g/s/user-profile/0000-000000-000000-0000-000000`.
        """
        return self.request.handler(
            method = method,
            url = url,
            data = data,
            is_login_required = is_login_required
        )


    def fetch_user(self, userId: str) -> UserProfile:
        """
        Fetches a user's profile.
        
        :param userId: The ID of the user to fetch.
        :type userId: str
        :return: The user's profile.
        :rtype: UserProfile
        
        This method is used to fetch a user's profile.
        
        Example usage:
        
        >>> x = client.fetch_user("0000-000000-000000-0000-000000")
        >>> print(x.nickname)
        "Example"
        >>> print(x.content)
        "This is an example profile."
        >>> print(x.icon)
        """
        return UserProfile(self.make_request(
            method = "GET",
            url = f"/g/s/user-profile/{userId}"
        ))
    

    def edit_profile(
        self,
        nickname: str = None,
        content: str = None,
        icon: str = None,
        backgroundColor: str = None,
        backgroundImage: str = None,
        defaultBubbleId: str = None
        ) -> UserProfile:
        """
        Edits the user's profile.

        :param nickname: The new nickname for the user.
        :type nickname: str
        :param content: The new content for the user's profile.
        :type content: str
        :param icon: The new icon image file for the user.
        :type icon: str
        :param backgroundColor: The new background color for the user's profile.
        :type backgroundColor: str
        :param backgroundImage: The new background image file for the user's profile.
        :type backgroundImage: str
        :param defaultBubbleId: The ID of the default bubble for the user's profile.
        :type defaultBubbleId: str
        :return: The response from the account's `edit_profile` method.
        :rtype: Response

        This method allows the authenticated user to edit their profile settings. Different aspects of the profile can be modified,
        such as the nickname, content, icon, background color, background image, and default bubble. Only the specified parameters will
        be updated. The `userId` parameter is set to the authenticated user's ID automatically.

        **Example usage:**

        To change the nickname and icon for the user:

        >>> response = client.edit_profile(nickname="New Nickname", icon="path/to/icon.jpg")
        ... if response.status == 200:
        ...     print("Profile edited successfully!")
        ... else:
        ...     print("Failed to edit profile.")
        """
        data = {
                "address": None,
                "latitude": 0,
                "longitude": 0,
                "mediaList": None,
                "eventSource": "UserProfileView",
                "timestamp": int(time() * 1000),
        }

        if nickname: data['nickname'] = nickname
        if icon: data['icon'] = self.upload_image(icon)
        if content: data['content'] = content
        if backgroundColor:
            data["extensions"] = {
                "style": {
                    "backgroundColor": backgroundColor
                    if backgroundColor.startswith("#")
                    else f"#{backgroundColor}"
                }
            }

        if backgroundImage:
            data["extensions"] = {
                "style": {
                    "backgroundMediaList": [
                        [100, self.upload_image(backgroundImage), None, None, None]
                    ]
                }
            }
        if defaultBubbleId:
            data["extensions"] = {"defaultBubbleId": defaultBubbleId}

        return UserProfile(self.make_request(
            method = "POST",
            url = f"/g/s/user-profile/{self.userId}",
            data = data
        ))


    @authenticated
    def send_message(self, content: str, chatId: str, **kwargs) -> CMessage:
        """
        Sends a message to a chat thread.

        :param content: The content of the message.
        :type content: str
        :param chatId: The ID of the chat thread to send the message to.
        :type chatId: str
        :param **kwargs: Additional parameters for the message.
        :return: A `CMessage` object containing the details of the sent message.
        :rtype: CMessage

        This method sends a message to a chat thread. The content and chat ID parameters are required. Additional parameters
        can be passed as keyword arguments. The method calls the `PrepareMessage` object's `json` method to prepare the message
        data. The result is a `CMessage` object containing the details of the sent message.
        """
        return CMessage(self.make_request(
            method="POST",
            url=f"/g/s/chat/thread/{chatId}/message",
            data = PrepareMessage(content=content, **kwargs).json()
            ))


    @authenticated
    def edit_profile(
        self,
        nickname: str = None,
        content: str = None,
        icon: str = None,
        backgroundColor: str = None,
        backgroundImage: str = None,
        defaultBubbleId: str = None
        ) -> UserProfile:
        """
        Edits the user's profile.

        :param nickname: The new nickname for the user.
        :type nickname: str
        :param content: The new content for the user's profile.
        :type content: str
        :param icon: The new icon image file for the user.
        :type icon: str
        :param backgroundColor: The new background color for the user's profile.
        :type backgroundColor: str
        :param backgroundImage: The new background image file for the user's profile.
        :type backgroundImage: str
        :param defaultBubbleId: The ID of the default bubble for the user's profile.
        :type defaultBubbleId: str
        :return: The response from the account's `edit_profile` method.
        :rtype: Response

        This method allows the authenticated user to edit their profile settings. Different aspects of the profile can be modified,
        such as the nickname, content, icon, background color, background image, and default bubble. Only the specified parameters will
        be updated. The `userId` parameter is set to the authenticated user's ID automatically.

        **Example usage:**

        To change the nickname and icon for the user:

        >>> response = client.edit_profile(nickname="New Nickname", icon="path/to/icon.jpg")
        ... if response.status == 200:
        ...     print("Profile edited successfully!")
        ... else:
        ...     print("Failed to edit profile.")
        """
        data = {
                "address": None,
                "latitude": 0,
                "longitude": 0,
                "mediaList": None,
                "eventSource": "UserProfileView",
                "timestamp": int(time() * 1000),
        }

        if nickname: data['nickname'] = nickname
        if icon: data['icon'] = self.upload_image(icon)
        if content: data['content'] = content
        if backgroundColor:
            data["extensions"] = {
                "style": {
                    "backgroundColor": backgroundColor
                    if backgroundColor.startswith("#")
                    else f"#{backgroundColor}"
                }
            }

        if backgroundImage:
            data["extensions"] = {
                "style": {
                    "backgroundMediaList": [
                        [100, self.upload_image(backgroundImage), None, None, None]
                    ]
                }
            }
        if defaultBubbleId:
            data["extensions"] = {"defaultBubbleId": defaultBubbleId}

        return UserProfile(self.make_request(
            method = "POST", url = f"/g/s/user-profile/{self.userId}",
            data = data
        ))


    @authenticated
    def start_chat(
        self,
        userId: Union[str, list],
        message: str,
        title: str = None,
        content: str = None,
        isGlobal: bool = False,
        publishToGlobal: bool = False
        ) -> ChatThread:
        """
        Starts a chat thread.
        :param userId: The ID or list of IDs of the users to invite to the chat.
        :type userId: Union[str, list]
        :param message: The initial message content.
        :type message: str
        :param title: The title of the chat thread (optional).
        :type title: str, optional
        :param content: Additional content for the message (optional).
        :type content: str, optional
        :param isGlobal: Indicates if the chat is global (optional, default: False).
        :type isGlobal: bool, optional
        :param publishToGlobal: Indicates if the chat should be published globally (optional, default: False).
        :type publishToGlobal: bool, optional
        :return: A `ChatThread` object representing the created chat thread.
        :rtype: ChatThread
        """
        try:
            userIds = [userId] if isinstance(userId, str) else userId
        except Exception as e:
            raise ValueError("Incorrect type for userId. <--- userId can be only a string or a list.") from e

        data = dict(
            title = title,
            inviteeUids = userIds,
            initialMessageContent = message,
            content = content,
            timestamp = int(time() * 1000),
            publishToGlobal = 0
        )

        if isGlobal: data.update({"type": 2, "eventSource": "GlobalComposeMenu"})
        else: data["type"] = 0

        if publishToGlobal: data["publishToGlobal"] = 1

        return ChatThread(
            self.make_request(method="POST", url="/g/s/chat/thread", data=data)
        )


    @authenticated
    def blocker_users(self, start: int = 0, size: int = 25) -> List[str]:
        """
        Retrieves a list of users what are blocking the logged account.
        :param start: The index to start retrieving the list from (optional, default: 0).
        :type start: int, optional
        :param size: The number of users to retrieve (optional, default: 25).
        :type size: int, optional
        :return: A list of user IDs representing the blocker users.
        :rtype: list
        """
        return self.make_request(
            method = "GET",
            url = f"/g/s/block/full-list?start={start}&size={size}"
        )["blockerUidList"]


    def fetch_wall_comments(self, userId: str, sorting: str = "newest", start: int = 0, size: int = 25) -> CommentList:
        """
        Fetches wall comments for a user.

        :param userId: The ID of the user whose wall comments will be fetched.
        :type userId: str
        :param sorting: The sorting method for the comments. Options: "newest" (default), "oldest", "top".
        :type sorting: str, optional
        :param start: The starting index of the comments to fetch (pagination). Default is 0.
        :type start: int, optional
        :param size: The number of comments to fetch (pagination). Default is 25.
        :type size: int, optional
        :return: The list of wall comments for the specified user.
        :rtype: CommentList
        :raises ValueError: If an incorrect sorting method is provided.

        This method retrieves wall comments for a specific user. The comments can be sorted by "newest" (default), "oldest", or "top".
        The `start` parameter specifies the index of the first comment to fetch, while the `size` parameter determines the number of
        comments to retrieve. The comments are returned as a `CommentList` object.

        **Example usage:**

        To fetch the newest 25 wall comments for a user:

        >>> comments = client.fetch_wall_comments(userId="00000000-0000-0000-0000-000000000000")
        >>> for comment in comments:
        ...     print(comment.content)
        """
        if sorting.lower() == "newest": sorting = "newest"
        elif sorting.lower() == "oldest": sorting = "oldest"
        elif sorting.lower() == "top": sorting = "vote"
        else: raise ValueError("Incorrect sorting method.")

        return CommentList(self.make_request(
            method = "GET",
            url = f"/g/s/user-profile/{userId}/g-comment?sort={sorting}&start={start}&size={size}"
        ))


    @authenticated
    def delete_message(self, chatId: str, messageId: str) -> ApiResponse:
        """
        Deletes a message in a chat thread.

        :param chatId: The ID of the chat thread where the message is located.
        :type chatId: str
        :param messageId: The ID of the message to delete.
        :type messageId: str
        :return: The API response indicating the success or failure of the deletion.
        :rtype: ApiResponse

        This method allows the authenticated user to delete a specific message in a chat thread. The `chatId` parameter identifies the
        chat thread, while the `messageId` parameter specifies the ID of the message to be deleted.

        **Note:** Only authorized users can delete messages in a chat thread. If the deletion is successful, the API response will indicate
        a successful deletion. Otherwise, an error message will be returned.

        **Example usage:**

        To delete a message with the ID "00000000-0000-0000-0000-000000000000" in a chat thread with the ID "00000000-0000-0000-0000-000000000000":

        >>> response = client.delete_message(chatId="00000000-0000-0000-0000-000000000000", messageId="00000000-0000-0000-0000-000000000000")
        >>> if response.success:
        ...     print("Message deleted successfully!")
        ... else:
        ...     print("Failed to delete message.")
        """
        return ApiResponse(self.make_request(
            method = "DELETE",
            url = f"/g/s/chat/thread/{chatId}/message/{messageId}"
        ))


    @authenticated
    def edit_chat(
        self,
        chatId: str,
        doNotDisturb: bool = None,
        pinChat: bool = None,
        title: str = None,
        icon: str = None,
        backgroundImage: str = None,
        content: str = None,
        announcement: str = None,
        coHosts: list = None,
        keywords: list = None,
        pinAnnouncement: bool = None,
        publishToGlobal: bool = None,
        canTip: bool = None,
        viewOnly: bool = None,
        canInvite: bool = None,
        fansOnly: bool = None
        ) -> list:
        """
        Edits the settings of a chat.

        :param chatId: The ID of the chat to be edited.
        :type chatId: str
        :param doNotDisturb: Set to True to enable "Do Not Disturb" mode for the chat, False to disable it. Default is None.
        :type doNotDisturb: bool, optional
        :param pinChat: Set to True to pin the chat, False to unpin it. Default is None.
        :type pinChat: bool, optional
        :param title: The new title for the chat.
        :type title: str, optional
        :param icon: The new icon image file for the chat.
        :type icon: str, optional
        :param backgroundImage: The new background image file for the chat.
        :type backgroundImage: str, optional
        :param content: The new content for the chat.
        :type content: str, optional
        :param announcement: The new announcement for the chat.
        :type announcement: str, optional
        :param coHosts: A list of user IDs to set as co-hosts for the chat.
        :type coHosts: list, optional
        :param keywords: A list of keywords to associate with the chat.
        :type keywords: list, optional
        :param pinAnnouncement: Set to True to pin the announcement, False to unpin it. Default is None.
        :type pinAnnouncement: bool, optional
        :param publishToGlobal: Set to True to publish the chat to the global feed, False to unpublish it. Default is None.
        :type publishToGlobal: bool, optional
        :param canTip: Set to True to enable tipping permissions for the chat, False to disable it. Default is None.
        :type canTip: bool, optional
        :param viewOnly: Set to True to enable view-only mode for the chat, False to disable it. Default is None.
        :type viewOnly: bool, optional
        :param canInvite: Set to True to allow members to invite others to the chat, False to disable it. Default is None.
        :type canInvite: bool, optional
        :param fansOnly: Set to True to enable "Fans Only" mode for the chat, False to disable it. Default is None.
        :type fansOnly: bool, optional
        :return: A list of HTTP status codes for each operation performed during the chat edit.
        :rtype: list

        This method allows the authenticated user to edit various settings of a chat, such as "Do Not Disturb" mode, pinning,
        title, icon, background image, content, announcement, co-hosts, keywords, pinning the announcement, publishing to the
        global feed, tipping permissions, view-only mode, allowing members to invite others, and "Fans Only" mode. Only the
        specified parameters will be updated.

        The function returns a list of HTTP status codes for each operation performed during the chat edit.

        **Example usage:**

        To edit the title and pin the chat:

        >>> response_codes = client.edit_chat(chatId="chat123", title="New Chat Title", pinChat=True)
        >>> if all(code == 200 for code in response_codes):
        ...     print("Chat edited successfully!")
        ... else:
        ...     print("Failed to edit chat.")
        """
        data = {"timestamp": int(time() * 1000)}

        if title: data["title"] = title
        if content: data["content"] = content
        if icon: data["icon"] = self.upload_image(icon)
        if keywords: data["keywords"] = keywords
        if announcement: data["extensions"] = {"announcement": announcement}
        if pinAnnouncement: data["extensions"] = {"pinAnnouncement": pinAnnouncement}
        if fansOnly: data["extensions"] = {"fansOnly": fansOnly}
        if publishToGlobal: data["publishToGlobal"] = 0
        if not publishToGlobal: data["publishToGlobal"] = 1

        responses = []

        if doNotDisturb is not None:
            responses.append(ApiResponse(self.make_request(
                method = "POST",
                url = f"/g/s/chat/thread/{chatId}/member/{self.userId}/alert",
                data = {
                    "alertOption": 2 if doNotDisturb else 1,
                    "timestamp": int(time() * 1000)
                }
            )).status_code)
        
        if pinChat is not None:
            responses.append(ApiResponse(self.make_request(
                method = "POST",
                url = f"/g/s/chat/thread/{chatId}/{'pin' if pinChat else 'unpin'}"
            )).status_code)
        
        if backgroundImage is not None:
            responses.append(ApiResponse(self.make_request(
                method = "POST",
                url = f"/g/s/chat/thread/{chatId}/member/{self.userId}/background",
                data = {
                    "media": [100, self.upload_image(backgroundImage), None],
                    "timestamp": int(time() * 1000)
                }
            )).status_code)
        
        if coHosts is not None:
            responses.append(ApiResponse(self.make_request(
                method = "POST",
                url = f"/g/s/chat/thread/{chatId}/co-host",
                data = {
                    "uidList": coHosts,
                    "timestamp": int(time() * 1000)
                }
            )).status_code)
        
        if viewOnly is not None:
            responses.append(ApiResponse(self.make_request(
                method = "POST",
                url = f"/g/s/chat/thread/{chatId}/view-only/{'enable' if viewOnly else 'disable'}"
            )).status_code)
        
        if canInvite is not None:
            responses.append(ApiResponse(self.make_request(
                method = "POST",
                url = f"/g/s/chat/thread/{chatId}/members-can-invite/{'enable' if canInvite else 'disable'}"
            )).status_code)
        
        if canTip is not None:
            responses.append(ApiResponse(
                method = "POST",
                url = f"/g/s/chat/thread/{chatId}/tipping-perm-status/{'enable' if canTip else 'disable'}"
            ).status_code)
        
        responses.append(ApiResponse(self.make_request(
            method = "POST",
            url = f"/g/s/chat/thread/{chatId}",
            data = data
        )).status_code)

        return responses


    @authenticated
    def follow(self, userId: Union[str, list]) -> ApiResponse:
        """
        Follows a user or a list of users.

        :param userId: The ID of the user or a list of user IDs to follow.
        :type userId: Union[str, list]
        :return: The status code of the API response.
        :rtype: int

        This method allows the authenticated user to follow another user or a list of users. The `userId` parameter can be a single
        user ID (string) or a list of user IDs. If a single user ID is provided, the method will follow that user. If a list of user IDs
        is provided, the method will follow all the users in the list.

        If a single user ID is provided, the function will make a POST request to "/g/s/user-profile/{userId}/member" to follow the user.
        If a list of user IDs is provided, the function will make a POST request to "/g/s/user-profile/{self.userId}/joined" with the
        `targetUidList` parameter set to the list of user IDs and the `timestamp` parameter set to the current timestamp.

        The function returns the status code of the API response.

        **Example usage:**

        To follow a single user:

        >>> response_code = client.follow(userId="user123")
        >>> if response_code == 200:
        ...     print("User followed successfully!")
        ... else:
        ...     print("Failed to follow user.")

        To follow multiple users:

        >>> user_ids = ["user123", "user456", "user789"]
        >>> response_code = client.follow(userId=user_ids)
        >>> if response_code == 200:
        ...     print("Users followed successfully!")
        ... else:
        ...     print("Failed to follow users.")
        """
        if isinstance(userId, str):
            return ApiResponse(self.make_request(
                method = "POST",
                url = f"/g/s/user-profile/{userId}/member"
            ))
        if isinstance(userId, list):
            return ApiResponse(self.make_request(
                method = "POST",
                url = f"/g/s/user-profile/{self.userId}/joined",
                data = {
                    "targetUidList": userId,
                    "timestamp": int(time() * 1000)
                }
            ))


    @authenticated
    def fetch_chats(self, start: int = 0, size: int = 25) -> ChatThreadList:
        """
        Fetches the chat threads for the authenticated user.

        :param start: The starting index of the chat threads to fetch. (Default: 0)
        :type start: int, optional
        :param size: The number of chat threads to fetch. (Default: 25)
        :type size: int, optional
        :return: A `ChatThreadList` object containing the fetched chat threads.
        :rtype: ChatThreadList

        This function sends a GET request to the API to fetch the chat threads that the authenticated user has joined.

        `ChatThreadList` represents a list of chat threads.

        **Example usage:**

        >>> threads = client.fetch_chats(start=0, size=10)
        ... for thread in threads:
        ...     print(thread.title)
        """
        return ChatThreadList(self.make_request(
            method = "GET",
            url = f"/g/s/chat/thread?type=joined-me&start={start}&size={size}"
        ))


    @authenticated
    def fetch_chat(self, chatId: str) -> ChatThread:
        """
        Fetches a chat thread by its ID.

        :param chatId: The ID of the chat thread to fetch.
        :type chatId: str
        :return: A `ChatThread` object representing the fetched chat thread.
        :rtype: ChatThread

        This function sends a GET request to the API to fetch a chat thread based on its ID.

        `ChatThread` represents a thread of messages in a chat.

        **Example usage:**

        >>> chat = client.fetch_chat("123456789")
        ... print(chat.thread_id)
        ... print(chat.messages)
        """
        return ChatThread(self.make_request(
            method = "GET",
            url = f"/g/s/chat/thread/{chatId}"
        ))


    @authenticated
    def fetch_chat_users(self, chatId: str, start: int = 0, size: int = 25) -> CChatMembers:
        """
        Fetches the users in a chat.

        :param chatId: The ID of the chat.
        :type chatId: str
        :param start: The start index for fetching the users. (Default: 0)
        :type start: int, optional
        :param size: The number of users to fetch. (Default: 25)
        :type size: int, optional
        :return: A `CChatMembers` object containing the chat members.
        :rtype: CChatMembers

        This function retrieves the users who are members of the specified chat.

        `CChatMembers` represents the members of a chat.

        **Example usage:**

        >>> chat_members = client.fetch_chat_users("chat123")
        ... for member in chat_members.nickname:
        ...     print(member)
        """
        return CChatMembers(self.make_request(
            method = "GET",
            url = f"/g/s/chat/thread/{chatId}/member?start={start}&size={size}&type=default&cv=1.2"
        ))


    @authenticated
    def invite_to_chat(self, chatId: str, userId: Union[str, list]) -> ApiResponse:
        """
        Invites user(s) to a chat.

        :param chatId: The ID of the chat to invite users to.
        :type chatId: str
        :param userId: The ID(s) of the user(s) to invite. It can be a string or a list of strings.
        :type userId: Union[str, list]
        :return: An `ApiResponse` object representing the response of the invitation request.
        :rtype: ApiResponse
        :raises TypeError: If userId is neither a string nor a list of strings.

        This function sends a POST request to invite user(s) to a chat.

        The user(s) specified by userId will be invited to join the chat identified by chatId.

        `ApiResponse` represents the response of an API request.

        **Example usage:**

        >>> response = client.invite_to_chat("chat123", "user456")
        ... print(response.status_code)
        """
        if isinstance(userId, str): userIds = [userId]
        elif isinstance(userId, list): userIds = userId
        else: raise TypeError("UserId must be a string or a list of strings.")

        return ApiResponse(self.make_request(
            method = "POST",
            url = f"/g/s/chat/thread/{chatId}/member/invite",
            data = {
                "timestamp": int(time() * 1000),
                "uids": userIds
            }
        ))


    @authenticated
    def kick(self, chatId: str, userId: str, allowRejoin: bool = True) -> ApiResponse:
        """
        Kicks a user from a chat.

        :param chatId: The ID of the chat.
        :type chatId: str
        :param userId: The ID of the user to be kicked.
        :type userId: str
        :param allowRejoin: Whether to allow the user to rejoin the chat. (Default: True)
        :type allowRejoin: bool, optional
        :return: An `ApiResponse` object representing the API response.
        :rtype: ApiResponse

        This function sends a DELETE request to the API to kick the specified user from the chat.

        `ApiResponse` represents the response from the API.

        **Example usage:**

        >>> response = client.kick("chat123", "user456")
        ... print(response.status_code)
        ... print(response.json())
        """
        return ApiResponse(self.make_request(
            method = "DELETE",
            url = f"/g/s/chat/thread/{chatId}/member/{userId}?allowRejoin={allowRejoin}"
        ))


    @authenticated
    def fetch_messages(self, chatId: str, size: int = 25, pageToken: str = None) -> CMessages:
        """
        Fetches messages from a chat.

        :param chatId: The ID of the chat to fetch messages from.
        :type chatId: str
        :param size: The number of messages to fetch. (Default: 25)
        :type size: int, optional
        :param pageToken: The page token for pagination. (Optional)
        :type pageToken: str, optional
        :return: A `CMessages` object representing the fetched messages.
        :rtype: CMessages

        This function retrieves messages from the specified chat using the provided parameters.

        `CMessages` represents a collection of messages in the chat.

        **Example usage:**

        >>> messages = client.fetch_messages("chat123", size=50, pageToken="token123")
        ... for message in messages:
        ...     print(message.text)
        """
        if pageToken is not None:
            return CMessages(self.make_request(
                method = "GET",
                url = f"/g/s/chat/thread/{chatId}/message?v=2&pagingType=t&pageToken={pageToken}&size={size}"
            ))
        return CMessages(self.make_request(
            method = "GET",
            url = f"/g/s/chat/thread/{chatId}/message?v=2&pagingType=t&size={size}"
        ))


    @authenticated
    def fetch_message(self, chatId: str, messageId: str) -> Message:
        """
        Fetches a specific message from a chat thread.

        :param chatId: The ID of the chat thread.
        :type chatId: str
        :param messageId: The ID of the message to fetch.
        :type messageId: str
        :return: A `Message` object representing the fetched message.
        :rtype: Message

        This function retrieves a specific message with the given `messageId` from the chat thread specified by `chatId`.

        `Message` represents a message in a chat thread.

        **Example usage:**

        >>> message = client.fetch_message("chat123", "message456")
        ... print(message.text)
        ... print(message.timestamp)
        """
        return Message(self.make_request(
            method = "GET",
            url = f"/g/s/chat/thread/{chatId}/message/{messageId}"
        ))


    @authenticated
    def search_community(self, aminoId: str) -> dict:
        """
        Search for a community by Amino ID or link.

        :param aminoId: The Amino ID or link of the community to search for.
        :type aminoId: str
        :return: A dictionary containing the search results.
        :rtype: dict

        This function sends a GET request to search for a community using the specified Amino ID or link.

        **Example usage:**

        >>> search_results = client.search_community("example_community").title
        ... print(search_results)
        """
        return self.make_request(
            method = "GET",
            url = f"/g/s/search/amino-id-and-link?q={aminoId}"
        )


    @authenticated
    def fetch_followers(self, userId: str, start: int = 0, size: int = 25) -> UserProfileList:
        """
        Fetches the followers of a user.

        :param userId: The ID of the user to fetch the followers for.
        :type userId: str
        :param start: The starting index of the followers list. (Default: 0)
        :type start: int, optional
        :param size: The number of followers to fetch. (Default: 25)
        :type size: int, optional
        :return: A `UserProfileList` object containing the fetched followers.
        :rtype: UserProfileList

        This function sends a GET request to the API to fetch the followers of a user.

        `UserProfileList` represents a list of user profiles.

        **Example usage:**

        >>> followers = client.fetch_followers("user123", start=0, size=10)
        ... for follower in followers.userId:
        ...     print(follower)
        """
        return UserProfileList(self.make_request(
            method = "GET",
            url = f"/g/s/user-profile/{userId}/member?start={start}&size={size}"
        ))


    def fetch_following(self, userId: str, start: int = 0, size: int = 25) -> UserProfileList:
        """
        Fetches the user profiles of the users that the specified user is following.

        :param userId: The ID of the user.
        :type userId: str
        :param start: The index to start fetching from. (Default: 0)
        :type start: int, optional
        :param size: The number of user profiles to fetch. (Default: 25)
        :type size: int, optional
        :return: A `UserProfileList` object containing the user profiles.
        :rtype: UserProfileList

        This function sends a GET request to the API to fetch the user profiles of the users that the specified user is following.

        `UserProfileList` represents a list of user profiles.

        **Example usage:**

        >>> following = client.fetch_following("user123")
        ... for profile in following.userId:
        ...     print(profile)
        """
        
        return UserProfileList(self.make_request(
            method = "GET",
            url = f"/g/s/user-profile/{userId}/joined?start={start}&size={size}"
        ))

    def large_fetch_following(self, userId: str, size: int = 25, pageToken: str = None, ignoreMembership: bool = True) -> FollowerList:
        """
        Fetches the user profiles of the users that the specified user is following.
        :param userId: The ID of the user.
        :type userId: str
        :param size: The number of user profiles to fetch. (Default: 25)
        :type size: int, optional
        :param pageToken: The page token to fetch from. (Default: None)
        :type pageToken: str, optional
        :param ignoreMembership: Whether to ignore membership. (Default: True)
        :type ignoreMembership: bool, optional
        :return: A `FollowerList` object containing the user profiles.
        :rtype: FollowerList
        
        This function sends a GET request to the API to fetch the user profiles of the users that the specified user is following.
        
        `FollowerList` represents a list of user profiles.
        
        **Example usage:**
    
        >>> following = client.large_fetch_following("user123")
        ... for userId in following.members.userId:
        ...     print(userId)
        """
        if pageToken:
            return FollowerList(self.make_request(
                method = "GET",
                url = f"/g/s/user-profile/{userId}/joined?size={size}&pageToken={pageToken}&pagingType=t&ignoreMembership={1 if ignoreMembership else 0}"
            ))

        return FollowerList(self.make_request(
            method = "GET",
            url = f"/g/s/user-profile/{userId}/joined?pagingType=t&size={size}&ignoreMembership={1 if ignoreMembership else 0}"
        ))

    def fetch_visitors(self, userId: str, start: int = 0, size: int = 25) -> UserProfileList:
        """
        Fetches the visitors of a user profile.

        :param userId: The ID of the user profile to fetch visitors for.
        :type userId: str
        :param start: The index of the first visitor to retrieve. (Default: 0)
        :type start: int, optional
        :param size: The maximum number of visitors to retrieve. (Default: 25)
        :type size: int, optional
        :return: A `UserProfileList` object containing the retrieved user profiles.
        :rtype: UserProfileList

        This function sends a GET request to the API to fetch the visitors of a user profile.

        `UserProfileList` represents a list of user profiles.

        **Example usage:**

        >>> visitors = client.fetch_visitors("123456", start=0, size=10)
        ... for visitor in visitors.userId:
        ...     print(visitor)
        """
        return UserProfileList(self.make_request(
            method = "GET",
            url = f"/g/s/user-profile/{userId}/visitors?start={start}&size={size}"
        ))


    @authenticated
    def blocked_users(self, start: int = 0, size: int = 25):
        """
        Retrieves a list of blocked users.

        :param start: The index of the first blocked user to retrieve. (Default: 0)
        :type start: int, optional
        :param size: The maximum number of blocked users to retrieve. (Default: 25)
        :type size: int, optional
        :return: A `UserProfileList` object containing the list of blocked users.
        :rtype: UserProfileList

        This function sends a GET request to the API to retrieve a list of blocked users.

        `UserProfileList` represents a list of user profiles.

        **Example usage:**

        >>> blocked_users = client.blocked_users(start=0, size=10)
        ... for user in blocked_users.userId:
        ...     print(user)
        """
        return UserProfileList(self.make_request(
            method = "GET",
            url = f"/g/s/block?start={start}&size={size}"
        ))


    @authenticated
    def mark_as_read(self, chatId: str, messageId: str) -> ApiResponse:
        """
        Marks a message as read.
        
        :param chatId: The ID of the chat to mark the message as read in.
        :type chatId: str
        :param messageId: The ID of the message to mark as read.
        :type messageId: str
        :return: An `ApiResponse` object containing the response from the API.
        :rtype: ApiResponse
        
        This function sends a POST request to the API to mark a message as read.
        
        `ApiResponse` represents a response from the API.
        
        **Example usage:**
        
        >>> x = client.mark_as_read("000000-0000-0000-000000", "000000-0000-0000-000000")
        ... print(x.status_code)
        """
        return ApiResponse(self.make_request(
            method = "POST",
            url = f"/g/s/chat/thread/{chatId}/mark-as-read",
            data = {
                "messageId": messageId,
                "timestamp": int(time() * 1000)
            }
        ))


    @authenticated
    def visit(self, userId: str) -> ApiResponse:
        """
        Visits a user profile.
        
        :param userId: The ID of the user profile to visit.
        :type userId: str
        :return: An `ApiResponse` object containing the response from the API.
        :rtype: ApiResponse
        
        This function sends a POST request to the API to visit a user profile.
        
        `ApiResponse` represents a response from the API.
        
        **Example usage:**
        
        >>> x = client.visit("000000-0000-0000-000000")
        ... print(x.status_code)
        """
        return ApiResponse(self.make_request(
            method = "POST",
            url = f"/g/s/user-profile/{userId}?action=visit"
        ))


    @authenticated
    def block(self, userId: str) -> ApiResponse:
        """
        Blocks a user.
        
        :param userId: The ID of the user to block.
        :type userId: str
        :return: An `ApiResponse` object containing the response from the API.
        :rtype: ApiResponse
        
        This function sends a POST request to the API to block a user.

        `ApiResponse` represents a response from the API.

        **Example usage:**

        >>> x = client.block("000000-0000-0000-000000")
        ... print(x.status_code)
        """
        return ApiResponse(self.make_request(
            method = "POST",
            url = f"/g/s/block/{userId}"
        ))


    @authenticated
    def unblock(self, userId: str) -> ApiResponse:
        """
        Unblocks a user.
        
        :param userId: The ID of the user to unblock.
        :type userId: str
        :return: An `ApiResponse` object containing the response from the API.
        :rtype: ApiResponse
        
        This function sends a DELETE request to the API to unblock a user.
        
        `ApiResponse` represents a response from the API.
        
        **Example usage:**
        
        >>> x = client.unblock("000000-0000-0000-000000")
        ... print(x.status_code)
        """
        return ApiResponse(self.make_request(
            method = "DELETE",
            url = f"/g/s/block/{userId}"
        ))


    @authenticated
    def join_request(self, comId: str, message: str = None):
        """
        Sends a join request to a community.
        
        :param comId: The ID of the community to send the join request to.
        :type comId: str
        :param message: The message to send with the join request.
        :type message: str, optional
        
        This function sends a POST request to the API to send a join request to a community.

        `ApiResponse` represents a response from the API.

        **Example usage:**

        >>> x = client.join_request("000000-0000-0000-000000", "Hello! I would like to join this community.")
        ... print(x.status_code)
        """
        return ApiResponse(self.make_request(
            method = "POST",
            url = f"/x{comId}/s/community/membership-request",
            data = {
                "message": message,
                "timestamp": int(time() * 1000)
            }
        ))


    @authenticated
    def flag_community(
        self,
        comId: str,
        reason: str,
        flagType: FlagTypes = FlagTypes.OFFTOPIC,
        isGuest: bool = False
        ) -> ApiResponse:
        """
        Flags a community.
        
        :param comId: The ID of the community to flag.
        :type comId: str
        :param reason: The reason for flagging the community.
        :type reason: str
        :param flagType: The type of flag to send.
        :type flagType: FlagTypes, optional
        :param isGuest: Whether or not the flag is from a guest account.
        :type isGuest: bool, optional
        
        This function sends a POST request to the API to flag a community.
        
        `ApiResponse` represents a response from the API.
        
        **Example usage:**
        
        >>> x = client.flag_community("000000-0000-0000-000000", "This community is offensive.", 1)
        ... print(x.status_code)
        """
        if reason is None: raise ValueError("Reason must be specified.")
        if flagType is None: raise ValueError("Flag type must be specified.")

        return ApiResponse(self.make_request(
            method = "POST",
            url = f"/x{comId}/s/item-flag",
            data = {
                "flagType": flagType.value if isinstance(flagType, FlagTypes) else flagType,
                "reason": reason,
                "timestamp": int(time() * 1000),
                "isGuest": "g-flag" if isGuest else "flag"
            }
        ))


    def fetch_linked_communities(self, userId: str) -> list:
        """
        Fetches a user's linked communities.
        
        :param userId: The ID of the user to fetch the linked communities of.
        :type userId: str
        :return: A list of linked communities.
        :rtype: list
        
        This function sends a GET request to the API to fetch a user's linked communities.
        
        **Example usage:**
        
        >>> x = client.fetch_linked_communities("000000-0000-0000-000000")
        ... print(x)
        """
        return self.make_request(
            method = "GET",
            url = f"/g/s/user-profile/{userId}/linked-community"
        )["linkedCommunityList"]


    def fetch_unlinked_communities(self, userId: str) -> list:
        """
        Fetches a user's unlinked communities.
        
        :param userId: The ID of the user to fetch the unlinked communities of.
        :type userId: str
        :return: A list of unlinked communities.
        :rtype: list
        
        This function sends a GET request to the API to fetch a user's unlinked communities.
        
        **Example usage:**
        
        >>> x = client.fetch_unlinked_communities("000000-0000-0000-000000")
        ... print(x)
        """
        return self.make_request(
            method = "GET",
            url = f"/g/s/user-profile/{userId}/unlinked-community"
        )["unlinkedCommunityList"]


    @authenticated
    def reorder_linked_communities(self, comIds: List[int]) -> ApiResponse:
        """
        Reorders a user's linked communities.
        
        :param comIds: A list of community IDs to reorder.
        :type comIds: List[int]
        :return: An `ApiResponse` object containing the response from the API.
        :rtype: ApiResponse
        
        This function sends a POST request to the API to reorder a user's linked communities.
        
        `ApiResponse` represents a response from the API.
        
        **Example usage:**
        
        >>> x = client.reorder_linked_communities([3, 2, 1])
        ... print(x.status_code)
        """
        return ApiResponse(self.make_request(
            method = "POST",
            url = f"/g/s/user-profile/{self.userId}/linked-community/reorder",
            data = {
                "timestamp": int(time() * 1000),
                "comIds": comIds
            }
        ))


    @authenticated
    def add_linked_community(self, comId: int) -> ApiResponse:
        """
        Adds a linked community to a user's profile.
        
        :param comId: The ID of the community to add.
        :type comId: int
        :return: An `ApiResponse` object containing the response from the API.
        :rtype: ApiResponse
        
        This function sends a POST request to the API to add a linked community to a user's profile.
        
        `ApiResponse` represents a response from the API.
        
        **Example usage:**
        
        >>> x = client.add_linked_community(3)
        ... print(x.status_code)
        """
        return ApiResponse(self.make_request(
            method = "POST",
            url = f"/g/s/user-profile/{self.userId}/linked-community/{comId}"
        ))


    @authenticated
    def remove_linked_community(self, comId: int) -> ApiResponse:
        """
        Removes a linked community from a user's profile.
        
        :param comId: The ID of the community to remove.
        :type comId: int
        :return: An `ApiResponse` object containing the response from the API.
        :rtype: ApiResponse
        
        This function sends a DELETE request to the API to remove a linked community from a user's profile.
        
        `ApiResponse` represents a response from the API.
        
        **Example usage:**
        
        >>> x = client.remove_linked_community(1)
        ... print(x.status_code)
        """
        return ApiResponse(self.make_request(
            method = "DELETE",
            url = f"/g/s/user-profile/{self.userId}/linked-community/{comId}"
        ))


    @authenticated
    def comment(self, message: str, userId: str = None, blogId: str = None, wikiId: str = None, replyTo: str = None) -> ApiResponse:
        """
        Comments on a user's profile, blog, or wiki.
        
        :param message: The message to comment.
        :type message: str
        :param userId: The ID of the user to comment on.
        :type userId: str
        :param blogId: The ID of the blog to comment on.
        :type blogId: str
        :param wikiId: The ID of the wiki to comment on.
        :type wikiId: str
        :param replyTo: The ID of the comment to reply to.
        :type replyTo: str
        :return: An `ApiResponse` object containing the response from the API.
        :rtype: ApiResponse
        
        This function sends a POST request to the API to comment on a user's profile, blog, or wiki.
        
        `ApiResponse` represents a response from the API.
        
        **Example usage:**
        
        >>> x = client.comment("Hello, world!", userId = "000000-0000-0000-000000")
        ... print(x.status_code)
        """
        if message is None: raise ValueError("Message must be specified.")

        data = {
            "content": message,
            "stickerId": None,
            "type": 0,
            "timestamp": int(time() * 1000)
        }

        if replyTo: data["respondTo"] = replyTo

        if userId:
            data["eventSource"] = "UserProfileView"
            return ApiResponse(self.make_request(
                method = "POST",
                url = f"/g/s/user-profile/{userId}/g-comment",
                data = data
            ))

        elif blogId:
            data["eventSource"] = "PostDetailView"
            return ApiResponse(self.make_request(
                method = "POST",
                url = f"/g/s/blog/{blogId}/g-comment",
                data = data
            ))
        
        elif wikiId:
            data["eventSource"] = "PostDetailView"
            url = f"/g/s/item/{wikiId}/g-comment"
            return ApiResponse(self.make_request(
                method = "POST",
                url = f"/g/s/item/{wikiId}/g-comment"
            ))
        
        else:
            raise ValueError("Either userId, blogId or wikiId must be specified.")


    @authenticated
    def delete_comment(self, commentId: str, userId: str = None, blogId: str = None, wikiId: str = None) -> ApiResponse:
        """
        Deletes a comment on a user's profile, blog, or wiki.
        
        :param commentId: The ID of the comment to delete.
        :type commentId: str
        :param userId: The ID of the user to delete the comment from.
        :type userId: str
        :param blogId: The ID of the blog to delete the comment from.
        :type blogId: str
        :param wikiId: The ID of the wiki to delete the comment from.
        :type wikiId: str
        :return: An `ApiResponse` object containing the response from the API.
        :rtype: ApiResponse
        
        This function sends a DELETE request to the API to delete a comment on a user's profile, blog, or wiki."""
        if userId:
            return ApiResponse(self.make_request(
                method = "DELETE",
                url = f"/g/s/user-profile/{userId}/g-comment/{commentId}"
            ))
        elif blogId:
            return ApiResponse(self.make_request(
                method = "DELETE",
                url = f"/g/s/blog/{blogId}/g-comment/{commentId}"
            ))
        elif wikiId:
            return ApiResponse(self.make_request(
                method = "DELETE",
                url = f"/g/s/item/{wikiId}/g-comment/{commentId}"
            ))
        else:
            raise ValueError("Either userId, blogId or wikiId must be specified.")


    @authenticated
    def like_blog(self, blogId: Union[str, list] = None, wikiId: str = None) -> ApiResponse:
        """
        Likes a blog or wiki.
        
        :param blogId: The ID of the blog to like.
        :type blogId: Union[str, list]
        :param wikiId: The ID of the wiki to like.
        :type wikiId: str
        :return: An `ApiResponse` object containing the response from the API.
        :rtype: ApiResponse
        
        This function sends a POST request to the API to like a blog or wiki.
        
        `ApiResponse` represents a response from the API.
        
        **Example usage:**
        
        >>> x = client.like_blog("000000-0000-0000-000000")
        ... print(x.status_code)
        """
        data = {"value": 4, "timestamp": int(time() * 1000)}

        if blogId:
            if isinstance(blogId, str):
                data["eventSource"] = "UserProfileView",
                return ApiResponse(self.make_request(
                    method = "POST",
                    url = f"/g/s/blog/{blogId}/g-vote?cv=1.2",
                    data = data
                ))
            elif isinstance(blogId, list):
                data["targetIdList"] = blogId,
                return ApiResponse(
                    self.make_request(
                        method = "POST", url="/g/s/feed/g-vote",
                        data = data
                    )
                )
            else: raise TypeError("blogId must be a string or a list.")

        elif wikiId:
            data["eventSource"] = "PostDetailView"
            return ApiResponse(self.make_request(
                method = "POST",
                url = f"/g/s/item/{wikiId}/g-vote?cv=1.2",
                data = data
            ))
        else:
            raise ValueError("Either blogId or wikiId must be specified.")


    @authenticated
    def unlike_blog(self, blogId: str = None, wikiId: str = None) -> ApiResponse:
        """
        Unlikes a blog or wiki.
        
        :param blogId: The ID of the blog to unlike.
        :type blogId: str
        :param wikiId: The ID of the wiki to unlike.
        :type wikiId: str
        :return: An `ApiResponse` object containing the response from the API.
        :rtype: ApiResponse
        
        This function sends a DELETE request to the API to unlike a blog or wiki.
        
        `ApiResponse` represents a response from the API.
        
        **Example usage:**
        
        >>> x = client.unlike_blog("000000-0000-0000-000000")
        ... print(x.status_code)
        """
        
        if blogId: return ApiResponse(self.make_request(
                method = "DELETE",
                url = f"{self.api}/g/s/blog/{blogId}/g-vote?eventSource=UserProfileView"
            ))
        elif wikiId: return ApiResponse(self.make_request(
                method = "DELETE",
                url = f"/g/s/item/{wikiId}/g-vote?eventSource=PostDetailView"
            ))
        else:
            raise ValueError("Either blogId or wikiId must be specified.")


    @authenticated
    def like_comment(self, commentId: str, userId: str = None, blogId: str = None, wikiId: str = None) -> ApiResponse:
        """
        Likes a comment on a user's profile, blog, or wiki.
        
        :param commentId: The ID of the comment to like.
        :type commentId: str
        :param userId: The ID of the user to like the comment on.
        :type userId: str
        :param blogId: The ID of the blog to like the comment on.
        :type blogId: str
        :param wikiId: The ID of the wiki to like the comment on.
        :type wikiId: str
        :return: An `ApiResponse` object containing the response from the API.
        :rtype: ApiResponse

        This function sends a POST request to the API to like a comment on a user's profile, blog, or wiki.

        `ApiResponse` represents a response from the API.

        **Example usage:**

        >>> x = client.like_comment("000000-0000-0000-000000", userId="000000-0000-0000-000000")
        ... print(x.status_code)
        """
        data = {"value": 4, "timestamp": int(time() * 1000)}

        if userId:
            data["eventSource"] = "UserProfileView"
            return ApiResponse(self.make_request(
                method = "POST",
                url = f"/g/s/user-profile/{userId}/comment/{commentId}/g-vote?cv=1.2&value=1",
                data = data
                ))
        
        elif blogId:
            data["eventSource"] = "PostDetailView"
            return ApiResponse(self.make_request(
                method = "POST",
                url = f"/g/s/blog/{blogId}/comment/{commentId}/g-vote?cv=1.2&value=1",
                data = data
            ))
        
        elif wikiId:
            data["eventSource"] = "PostDetailView"
            return ApiResponse(self.make_request(
                method = "POST",
                url = f"/g/s/item/{wikiId}/comment/{commentId}/g-vote?cv=1.2&value=1",
                data = data
            ))
        
        else: raise ValueError("Either userId, blogId or wikiId must be specified.")


    @authenticated
    def unlike_comment(self, commentId: str, userId: str = None, blogId: str = None, wikiId: str = None) -> ApiResponse:
        """
        Unlikes a comment on a user's profile, blog, or wiki.
        
        :param commentId: The ID of the comment to unlike.
        :type commentId: str
        :param userId: The ID of the user to unlike the comment on.
        :type userId: str
        :param blogId: The ID of the blog to unlike the comment on.
        :type blogId: str
        :param wikiId: The ID of the wiki to unlike the comment on.
        :type wikiId: str
        :return: An `ApiResponse` object containing the response from the API.
        :rtype: ApiResponse
        
        This function sends a DELETE request to the API to unlike a comment on a user's profile, blog, or wiki.
        
        `ApiResponse` represents a response from the API.
        
        **Example usage:**
        
        >>> x = client.unlike_comment("000000-0000-0000-000000", userId="000000-0000-0000-000000")
        ... print(x.status_code)
        """
        if userId:
            return ApiResponse(self.make_request(
                method = "DELETE",
                url = f"/g/s/user-profile/{userId}/comment/{commentId}/g-vote?eventSource=UserProfileView"
            ))
        elif blogId:
            return ApiResponse(self.make_request(
                method = "DELETE",
                url = f"/g/s/blog/{blogId}/comment/{commentId}/g-vote?eventSource=PostDetailView"
            ))
        elif wikiId:
            return ApiResponse(self.make_request(
                method = "DELETE",
                url = f"/g/s/item/{wikiId}/comment/{commentId}/g-vote?eventSource=PostDetailView"
            ))
        else: raise ValueError("Either userId, blogId or wikiId must be specified.")


    def fetch_supported_languages(self) -> List[str]:
        """
        Fetches a list of supported languages for the API.
        
        :return: A list of supported languages.
        :rtype: List[str]
        
        This function sends a GET request to the API to fetch a list of supported languages.
        
        **Example usage:**
        
        >>> x = client.fetch_supported_languages()
        ... print(x)
        """
        return self.make_request(
            method="GET",
            url="/g/s/community-collection/supported-languages?start=0&size=100",
        )["supportedLanguages"]


    @authenticated
    def fetch_ta_announcement(self, language: str = "en", start: int = 0, size: int = 25) -> dict:
        """
        Fetches the latest announcements from the API.

        :param language: The language to fetch the announcements in.
        :type language: str
        :param start: The index to start fetching announcements from.
        :type start: int
        :param size: The number of announcements to fetch.
        :type size: int
        :return: A dictionary containing the announcements.
        :rtype: dict

        This function sends a GET request to the API to fetch the latest announcements.

        **Example usage:**

        >>> x = client.fetch_ta_announcement()
        ... print(x)
        """
        if language not in self.fetch_supported_languages():
            raise ValueError("Invalid language.")
        
        return self.make_request(
            method = "GET",
            url = f"/g/s/announcement?language={language}&start={start}&size={size}"
        )


    @authenticated
    def join_chat(self, chatId: str) -> ApiResponse:
        """
        Joins the authenticated user to a chat thread.

        :param chatId: The ID of the chat thread to join.
        :type chatId: str
        :return: The API response.
        :rtype: ApiResponse

        This method joins the authenticated user to a chat thread. The user must be a member of the community that the chat
        thread belongs to in order to join the thread.

        **Note:** This method requires authentication. If the client is not authenticated, a `LoginRequired` exception will
        be raised.
        """
        return ApiResponse(self.make_request(
            method="POST",
            url=f"/g/s/chat/thread/{chatId}/member/{self.userId}"
            ))


    @authenticated
    def leave_chat(self, chatId: Union[str, List[str]]) -> ApiResponse:
        """
        Removes the authenticated user from a chat thread.

        :param chatId: A list of chat thread IDs to leave or a single chat thread ID to leave.
        :type chatId: Union[str, List[str]]
        :return: The API response.
        :rtype: ApiResponse

        This method removes the authenticated user from a chat thread. The user must be a member of the chat thread in order
        to leave it.

        **Note:** This method requires authentication. If the client is not authenticated, a `LoginRequired` exception will
        be raised.
        """
        return ApiResponse(self.make_request(
            method="DELETE",
            url = f"/g/s/chat/thread/leave?threadIds={','.join(chatId) if isinstance(chatId, list) else chatId}"
            ))


    @authenticated
    def join_community(self, community_id: int, invitationId = None) -> ApiResponse:
        """
        Joins the user to a community with the provided community ID.

        :param community_id: The ID of the community to join.
        :type community_id: int
        :param invitationId: The ID of the invitation link.
        :type invitationId: str
        :return: An ApiResponse object containing the server response.
        :rtype: ApiResponse
        :raises LoginRequired: If the user is not logged in.

        This function first checks if the client is logged in by checking for a valid session ID. If not, it raises a
        LoginRequired exception.

        If the client is logged in, the function sends a POST request to the server to join the specified community.
        The request includes the client's session ID and the ID of the community to join.

        The function returns an ApiResponse object containing the server response.

        **Note:** This function can be used to join the user to a community with the provided community ID. Once joined,
        the user can make API calls related to the community, such as posting or retrieving posts.
        """
        data = {"timestamp": int(time() * 1000)}
        if invitationId:
            data["invitationId"] = invitationId

        return ApiResponse(
            self.make_request(
                method="POST",
                url=f"/x{community_id}/s/community/join",
                data = data
        ))


    @authenticated
    def leave_community(self, community_id: int) -> ApiResponse:
        """
        Leaves the user from a community with the provided community ID.

        :param community_id: The ID of the community to leave.
        :type community_id: int
        :return: An ApiResponse object containing the server response.
        :rtype: ApiResponse
        :raises LoginRequired: If the user is not logged in.

        This function first checks if the client is logged in by checking for a valid session ID. If not, it raises a
        LoginRequired exception.

        If the client is logged in, the function sends a POST request to the server to leave the specified community.
        The request includes the client's session ID and the ID of the community to leave.

        The function returns an ApiResponse object containing the server response.

        **Note:** This function can be used to leave the user from a community with the provided community ID. Once left,
        the user will no longer have access to the community and will not be able to make API calls related to the community.
        """
        return ApiResponse(self.make_request(
            method="POST",
            url=f"/x{community_id}/s/community/leave"
            ))
    

    def fetch_community(self, community_id: int) -> CCommunity:
        """
        Fetches the community information for the community with the provided community ID.

        :param community_id: The ID of the community to fetch.
        :type community_id: int
        :return: A CCommunity object representing the community information.
        :rtype: CCommunity

        This function sends a GET request to the server to fetch the community information for the community with the
        provided community ID. The request includes the ID of the community to fetch.

        The function returns a CCommunity object representing the community information. The CCommunity object contains
        attributes such as the members count, the community's layout, and other community information.

        **Note:** This function can be used to fetch the community information for a community on Amino. The community
        information can be used to display information about the community such as the community's name, description, and
        other information.
        """
        return CCommunity(self.make_request(
            method="GET",
            url=f"/g/s-x{community_id}/community/info"
            ))
    

    @authenticated
    def joined_communities(self) -> CCommunityList:
        """
        Retrieves the list of communities that the authenticated user has joined.

        :return: The list of communities.
        :rtype: CCommunityList

        This method returns the list of communities that the authenticated user has joined. The list includes information
        such as the community ID, name, description, and theme.

        **Note:** This method requires authentication. If the client is not authenticated, a `LoginRequired` exception will
        be raised.
        """
        return CCommunityList(self.make_request(
            method="GET",
            url="/g/s/community/joined"
            ))


    def fetch_object_id(self, link: str) -> str:
        """
        Fetches the object ID given a link to the object.

        :param link: The link to the object.
        :type link: str
        :raises NotLoggedIn: If the user is not logged in.
        :raises MissingCommunityId: If the community ID is missing.
        :return: The ID of the object.
        :rtype: str

        The `community` decorator is used to ensure that the user is logged in and the community ID is present.
        The method caches the object ID for faster access in future calls.

        **Example usage:**

        >>> object_id = client.community.fetch_object_id(link="https://aminoapps.com/p/w2Fs6H")
        >>> print(object_id)
        """

        KEY = str((link, "OBJECT_ID"))
        if not self.cache.get(KEY):
            self.cache.set(KEY, self.make_request(
                method = "GET",
                url = f"/g/s/link-resolution?q={link}"
                ))
        return LinkInfo(self.cache.get(KEY)).objectId


    def fetch_object_info(self, link: str) -> LinkInfo:
        """
        Fetches information about an object given its link.

        :param link: The link to the object.
        :type link: str
        :raises NotLoggedIn: If the user is not logged in.
        :raises MissingCommunityId: If the community ID is missing.
        :return: A LinkInfo object containing information about the object.
        :rtype: LinkInfo

        The `community` decorator is used to ensure that the user is logged in and the community ID is present.
        The method caches the object information for faster access in future calls.

        `LinkInfo`:

        - `data`: The raw response data from the API.
        - `linkInfoV2`: The link information data.
        - `path`: The path of the object.
        - `extensions`: The extensions data.
        - `objectId`: The ID of the object.
        - `shareURLShortCode`: The short code of the share URL.
        - `targetCode`: The target code.
        - `ndcId`: The NDC ID.
        - `comId`: The community ID.
        - `fullPath`: The full path of the object.
        - `shortCode`: The short code of the object.
        - `objectType`: The type of the object.

        **Example usage:**

        >>> object_info = client.community.fetch_object_info(link="https://aminoapps.com/p/w2Fs6H")
        >>> print(object_info.objectId)
        """

        KEY = str((link, "OBJECT_INFO"))
        if not self.cache.get(KEY):
            self.cache.set(KEY, self.make_request(
                method = "GET",
                url = f"/g/s/link-resolution?q={link}"
                ))
        return LinkInfo(self.cache.get(KEY))

    def fetch_public_communities(self, type: str = "discover") -> CCommunityList:
        """
        Fetches a list of public communities.

        This method retrieves a list of public communities based on the specified parameters.

        :param type: The type of communities to fetch (default is "discover").
        :type type: str
        :return: A `CCommunityList` object containing the list of public communities.
        :rtype: CCommunityList
        
        **example usage:**
        
        >>> communities = client.fetch_public_communities()
        >>> for community in communities.name:
        >>>     print(community)
        """
        return CCommunityList(self.make_request(
            method = "GET",
            url = f"/g/s/topic/0/feed/community?type={type}&categoryKey=recommendation&moduleId=0c56a709-1f96-474d-ae2f-4225d0e998e5"
        ))
    
    def fetch_available_communities(self, start: int = 0, size: int = 25, language: str = "en") -> CCommunityList:
        """
        Fetches a list of available communities.

        This method retrieves a list of communities that are available for exploration.

        :param start: The starting index of the communities to fetch (default is 0).
        :type start: int
        :param size: The number of communities to fetch in a single request (default is 25).
        :type size: int
        :param language: The language code used for filtering communities (default is "en").
        :type language: str
        :return: A CCommunityList object containing information about the available communities.
        :rtype: CCommunityList
        :raises: Any exceptions raised during the API request process.

        The `CCommunityList` object provides access to information about multiple communities:

        - `json`: The raw response data from the API.
        - `comId`: The list of community Ids.
        - `name`: The list of community names.

        **Example usage:**

        >>> community_list = client.community.fetch_available_communities(start=0, size=10, language="en")
        >>> for name, comId in zip(community_list.name, community_list.comId):
        >>>     print(name, comId])
        """
        return CCommunityList(self.make_request(
            method = "GET",
            url = f"/g/s/topic/0/feed/community?language={language}&type=web-explore&categoryKey=recommendation&start={start}&size={size}&pagingType=t"
        ))

    @authenticated
    def unfollow(self, userId: str) -> ApiResponse:
        """
        Unfollows a user.
        :param userId: The ID of the user to unfollow.
        :type userId: str
        :return: An ApiResponse object containing the response data from the API.
        :rtype: ApiResponse
        This function allows the logged-in user to unfollow another user specified by their ID.
        After successful execution, the user will no longer be following the specified user.
        **Example usage:**
        >>> response = client.unfollow(userId="user123")
        >>> print(response.status_code)
        """
        return ApiResponse(self.make_request(
            method = "DELETE",
            url = f"/g/s/user-profile/{userId}/member/{self.userId}"
        ))
    
    @authenticated
    def fetch_notifications(self, start: int = 0, size: int = 25):
        """
        Fetches the notifications for the authenticated user.
        :param start: The starting index of the notifications to fetch (default is 0).
        :type start: int
        :param size: The number of notifications to fetch in a single request (default is 25).
        :type size: int
        :return: A list of notifications.
        :rtype: list
        This function allows the logged-in user to fetch their notifications.
        **Example usage:**
        >>> notifications = client.fetch_notifications(start=0, size=25)
        >>> for notification in notifications:
        >>>     print(notification)
        """
        return GlobalNotificationList(self.make_request(
            method = "GET",
            url = f"/g/s/notification?start={start}&size={size}"
        ))
    
    
    @authenticated
    def delete_notification(self, notificationId: str) -> ApiResponse:
        """
        Deletes a notification.
        :param notificationId: The ID of the notification to delete.
        :type notificationId: str
        :return: An ApiResponse object containing the response data from the API.
        :rtype: ApiResponse
        This function allows the logged-in user to delete a notification.
        **Example usage:**
        >>> response = client.delete_notification(notificationId="notification123")
        >>> print(response.status_code)
        """
        return ApiResponse(self.make_request(
            method = "DELETE",
            url = f"/g/s/notification/{notificationId}"
        ))


class Notification:
    def __init__(self, data: dict):
        self.data: dict = data

    @property
    def __parser__(self) -> dict:
        try:
            return self.data.get("o")
        except Exception:
            return self.data
    
    @property
    def payload(self) -> dict:
        """Returns the payload of the notification."""
        return self.__parser__.get("payload")
    
    @property
    def exp(self) -> int:
        """Returns the expiration date of the notification."""
        return self.payload.get("exp")
    
    @property
    def ndcId(self) -> int:
        """Returns the NDC ID of the notification."""
        return self.payload.get("ndcId")
    
    @property
    def comId(self) -> int:
        """Returns the COM ID of the notification."""
        return self.ndcId
    
    @property
    def chatId(self) -> str:
        """Returns the chat ID of the notification."""
        return self.payload.get("tid")
    
    @property
    def aps(self) -> dict:
        """Returns the APS of the notification."""
        return self.payload.get("aps")
    
    @property
    def sound(self) -> str:
        """Returns the sound of the notification."""
        return self.aps.get("sound")
    
    @property
    def alert(self) -> str:
        """Returns the alert of the notification."""
        return self.aps.get("alert")
    
    @property
    def notification_type(self) -> int:
        """Returns the notification type of the notification."""
        return self.payload.get("notifType")
    
    @property
    def id(self) -> str:
        """Returns the ID of the notification."""
        return self.payload.get("id")
    
    def json(self) -> dict:
        """Returns the JSON data of the notification."""
        return self.data


from requests import get
from functools import wraps
from diskcache import Cache
from threading import Thread
from base64 import b64encode
from contextlib import suppress
from colorama import Fore, Style
from time import sleep as delay, time
from inspect import signature as inspect_signature
from typing import BinaryIO, Callable, List, Union

from .entities import *
from .utilities.commands import Command, Commands

__all__ = (
    "Context",
    "EventHandler",
    "WaitForMessage",
    )

class WaitForMessage:
    def __init__(self, status_code: int) -> None:
        self.status_code = status_code
        self._MESSAGE_TIMEOUT = None
        self._INCORRECT_MESSAGE = None
        self._MESSAGE_FOUND = None

        self.__set_status_code__()

    @property
    def MESSAGE_TIMEOUT(self) -> bool:
        """Whether or not the timeout was reached."""
        return self._MESSAGE_TIMEOUT
    
    @MESSAGE_TIMEOUT.setter
    def MESSAGE_TIMEOUT(self, value: bool) -> None:
        self._MESSAGE_TIMEOUT = value

    @property
    def MESSAGE_NOT_FOUND(self) -> bool:
        """Whether or not the message was not found."""
        return self._MESSAGE_NOT_FOUND
    
    @MESSAGE_NOT_FOUND.setter
    def MESSAGE_NOT_FOUND(self, value: bool) -> None:
        self._MESSAGE_NOT_FOUND = value

    @property
    def MESSAGE_FOUND(self) -> bool:
        """Whether or not the message was found."""
        return self._MESSAGE_FOUND
    
    @MESSAGE_FOUND.setter
    def MESSAGE_FOUND(self, value: bool) -> None:
        self._MESSAGE_FOUND = value
    
    def __set_status_code__(self):
        if self.status_code == 200:
            self.MESSAGE_FOUND = True
            self.MESSAGE_NOT_FOUND = False
            self.MESSAGE_TIMEOUT = False

        elif self.status_code == 404:
            self.MESSAGE_FOUND = False
            self.MESSAGE_NOT_FOUND = True
            self.MESSAGE_TIMEOUT = False

        elif self.status_code == 500:
            self.MESSAGE_FOUND = False
            self.MESSAGE_NOT_FOUND = False
            self.MESSAGE_TIMEOUT = True

    def __repr__(self) -> str:
        return f"<WaitForMessage status_code={self.status_code} MESSAGE_TIMEOUT={self.MESSAGE_TIMEOUT} MESSAGE_NOT_FOUND={self.MESSAGE_NOT_FOUND} MESSAGE_FOUND={self.MESSAGE_FOUND}>"

class Context:
    """
    Context class that handles context.

    This class is used to handle context for commands.

    Special Attributes:
    __slots__ : tuple
        A tuple containing a fixed set of attributes to optimize memory usage.

    Attributes:
    message : Message
        The message object.
    bot : Bot
        The bot object.
    request : RequestHandler
        The request handler object.
    userId : str
        The user ID.
    intents : bool
        Whether or not intents are enabled.

"""
    __slots__ = (
        "message",
        "bot",
        "request",
        "userId",
        "intents"
    )
    def __init__(self, message: Message, bot):
        self.message:   Message = message 
        self.bot        = bot
        self.request    = self.bot.request
        self.userId:    str = self.request.userId
        self.intents:   bool = self.bot.intents


    @property
    def author(self) -> MessageAuthor:
        """The author of the message."""
        with suppress(AttributeError):
            return self.message.author

    @property
    def communityId(self) -> str:
        """Sets the url to community/global."""
        return {True: "g", False: f"x{self.message.comId}"}[self.message.comId == 0]

    @property
    def comId(self) -> int:
        """The community ID."""
        return self.message.comId

    @property
    def chatId(self) -> str:
        """The chat ID."""
        return self.message.chatId

    @property
    def api(self) -> str:
        """The API url."""
        return "https://service.aminoapps.com/api/v1"
    
    @property
    def cache(self) -> Cache:
        """The cache."""
        return Cache("cache")

    @property
    def __message_endpoint__(self) -> str:
        """The message endpoint."""
        return f"/{self.communityId}/s/chat/thread/{self.message.chatId}/message"


    def __typing__(func: Callable) -> Callable:
        @wraps(func)
        def wrapper(*args, **kwargs):
            Thread(target=args[0].__rt__, args=(args[0].comId, args[0].chatId)).start()
            return func(*args, **kwargs)
        return wrapper


    def _run(func: Callable) -> Callable:
        def wrapper(*args, **kwargs):
                if isinstance(args[0], Context):
                    return func(*args, **kwargs)
                else:
                    raise MustRunInContext
        return wrapper


    def __purge__(self, data: dict) -> dict:
        return {k: v for k, v in data.items() if v is not None}


    def __prepare_message__(self, **kwargs) -> dict:
        return self.__purge__(self.__parse_kwargs__(**kwargs))    


    def __read_image__(self, image: Union[str, BinaryIO]) -> BinaryIO:
        try:
            return get(image).content if image.startswith("http") else open(image, "rb").read()
        except InvalidImage as e:
            raise InvalidImage from e


    def __parse_kwargs__(self, **kwargs) -> dict:
        return {
            "content": kwargs.get("content"),
            "type": kwargs.get("type"),
            "mediaType": kwargs.get("mediaType"),
            "mediaValue": kwargs.get("mediaValue"),
            "mediaUploadValue": kwargs.get("mediaUploadValue"),
            "stickerId": kwargs.get("stickerId"),
            "attachedObject": kwargs.get("attachedObject"),
            "uid": self.userId
            } 


    def __message__(self, **kwargs) -> dict:
        return PrepareMessage(**kwargs).json()


    def __send_message__(self, **kwargs) -> CMessage:
        return CMessage(self.request.handler(
            method = "POST",
            url = self.__message_endpoint__,
            data = self.__message__(**kwargs)
            ))


    def __st__(self, comId: str, chatId: str):
        return self.bot.send_websocket_message({
            "o":{
                "actions":["Typing"],
                "target":f"ndc://x{comId}/chat-thread/{chatId}",
                "ndcId":comId,
                "params":{"topicIds":[],"threadType":2},
                "id":randint(0, 100)},
                "t":304
                })


    def __et__(self, comId: str, chatId: str):
        def wrapper():
            return self.bot.send_websocket_message({
                "o":{
                    "actions":["Typing"],
                    "target":f"ndc://x{comId}/chat-thread/{chatId}",
                    "ndcId":comId,
                    "params":{"duration":0,"topicIds":[],"threadType":2},
                    "id":randint(0, 100)},
                    "t":306
                    })
        delay(2.5)
        return wrapper()


    def __rt__(self, comId: str, chatId: str):
        self.__st__(comId, chatId)
        self.__et__(comId, chatId)


    def _delete(self, delete_message: CMessage, delete_after: int = 5) -> ApiResponse:
        """
        `delete` - Deletes a message.
        
        `**Parameters**`
        - `delete_message` - The message to delete.
        - `delete_after` - The time to delay before deleting the message.
        
        """
        delay(delete_after)
        return ApiResponse(self.request.handler(
            method = "DELETE",
            url = f"/{self.communityId}/s/chat/thread/{self.message.chatId}/message/{delete_message.messageId}"
            ))
    
    def wait_for_message(self, message: str, timeout: int = 10) -> WaitForMessage:
        """
        `wait_for_message` - This waits for a specific message within a certain timeout period. 
        
        `**Parameters**`
        - `message` : str
            The specific message to wait for in the cache.
        - `timeout` : int, optional
            The maximum time to wait for the message in seconds. Default is 10.
        
        `**Returns**`
        - `WaitForMessage`: The WaitForMessage object.

            - `MESSAGE_TIMEOUT` : bool
                - Whether or not the timeout was reached.
            - `MESSAGE_NOT_FOUND` : bool
                - Whether or not the message was not found.
            - `MESSAGE_FOUND` : bool
                - Whether or not the message was found.

        
        `**Example**`
        ```py
        @bot.on_member_join()
        def on_member_join(ctx: Context):
            if ctx.comId != bot.community.community_id:
                return
                
            TIMEOUT = 15

            ctx.send(content="Welcome to the chat! Please verify yourself by typing `$verify` in the chat.", delete_after=TIMEOUT)

            response = ctx.wait_for_message(message="$verify", timeout=15)

            if response.MESSAGE_TIMEOUT:
                ctx.send(content="You took too long to verify yourself. You have been kicked from the chat.", delete_after=TIMEOUT)
                return bot.community.kick(userId=ctx.author.userId, chatId=ctx.chatId, allowRejoin=True, comId=ctx.comId)

            elif response.MESSAGE_NOT_FOUND:
                ctx.send(content="Invalid verification code. You have been kicked from the chat.", delete_after=TIMEOUT)

            elif response.MESSAGE_FOUND:
                ctx.send(content="You have been verified!", delete_after=TIMEOUT)
        ```
        """
        if not self.intents:
            raise IntentsNotEnabled

        start = time()

        with self.bot.cache as cache:
            while time() - start < timeout:
                cached_message = cache.get(f"{self.message.chatId}_{self.message.author.userId}")

                if cached_message == message:
                    cache.pop(f"{self.message.chatId}_{self.message.author.userId}")
                    return WaitForMessage(status_code=200)

                if all([cached_message is not None, cached_message != message]):
                    cache.pop(f"{self.message.chatId}_{self.message.author.userId}")
                    return WaitForMessage(status_code=404)

            cache.pop(f"{self.message.chatId}_{self.message.author.userId}")
            return WaitForMessage(status_code=500)

    @_run
    @__typing__
    def send(self, content: str, delete_after: int= None, mentioned: Union[str, List[str]]= None) -> CMessage:
        """
        `send` - This sends a message.

        `**Parameters**``
        - `content` - The message you want to send.
        - `delete_after` - The time in seconds before the message is deleted. [Optional]
        - `mentioned` - The user(s) you want to mention. [Optional]

        `**Returns**`` - CMessage object.

        `**Example**``
        ```py
        @bot.on_text_message()
        def on_text_message(ctx: Context):
            ctx.send(content="Hello World!", delete_after=None)
        ```
        """
        message: CMessage = self.__send_message__(
            content=content,
            extensions = {
            "mentionedArray": [{"uid": user} for user in mentioned] if mentioned else None
            })

        Thread(target=self._delete, args=(message, delete_after)).start() if delete_after else None

        return message


    @_run
    @__typing__
    def reply(self, content: str, delete_after: int= None, mentioned: Union[str, List[str]]= None) -> CMessage:
        """
        `reply` - This replies to the message.

        `**Parameters**``
        - `content` - The message you want to send.
        - `delete_after` - The time in seconds before the message is deleted. [Optional]
        - `mentioned` - The user(s) you want to mention. [Optional]

        `**Returns**`` - CMessage object.

        `**Example**``
        ```py
        @bot.on_text_message()
        def on_text_message(ctx: Context):
            ctx.reply(content = "Hello World!", delete_after = None)
        ```
        """
        message: CMessage = self.__send_message__(
            content=content,
            replyMessageId=self.message.messageId,
            extensions = {
            "mentionedArray": [{"uid": user} for user in mentioned] if mentioned else None
            })
        
        Thread(target=self._delete, args=(message, delete_after)).start() if delete_after else None
        
        return message


    def prepare_mentions(self, mentioned: list) -> list:
        """
        `prepare_mentions` - This prepares the mentions for the message.
        
        `**Parameters**``
        - `mentioned` - `ctx.message.mentioned_user_names`.
        
        `**Returns**``
        - `list` - The list of mentions to use as your `message`

        `**Example**``
        ```py
        @bot.command("mention")
        def mention(ctx: Context):
            mentioned_users = ctx.message.mentioned_user_names
            if not mentioned_users:
                return ctx.reply("You didn't mention anyone!")

            mentioned = ctx.prepare_mentions(mentioned_users)
            return ctx.reply(
                "Mentioned: " + ", ".join(mentioned), mentioned=list(mentioned_users)
            )
        """
        return [f"\u200e\u200f@{username}\u202c\u202d" for username in mentioned]


    @_run
    @__typing__
    def send_link_snippet(self, image: str, message: str = "[c]", link: str = "ndc://user-me", mentioned: list = None) -> CMessage:
        """
        `send_link_snippet` - This sends a link snippet.

        `**Parameters**``
        - `image` - The image you want to send. Recommended size: 807x216
        - `message` - The message you want to send.
        - `link` - The link you want to send. [Optional]

        `**Returns**`` - CMessage object.

        `**Example**``
        ```py
        @bot.command("linksnippet")
        def linksnippet(ctx: Context):
            return ctx.send_link_snippet(
                image = "https://i.imgur.com/8ZQZ9Zm.png",
                message = "Hello World!",
                link = "https://www.google.com"
            )
        ```
        """
        if mentioned is None: mentioned = []

        message: CMessage = self.__send_message__(
            content=message,
            extensions = {
                "linkSnippetList": [{
                "mediaType": 100,
                "mediaUploadValue": self.encode_media(
                    self.__handle_media__(media=image, content_type="image/jpg", media_value=False)
                ),
                "mediaUploadValueContentType": "image/png",
                "link": link
                }],
            "mentionedArray": [{"uid": user} for user in mentioned] if mentioned else None
            })

        return message


    @_run
    @__typing__
    def send_embed(
        self,
        message: str,
        title: str,
        content: str,
        image: str,
        link: str = "ndc://user-me",
        mentioned: Union[str, List[str]]= None
        ) -> CMessage:
        """
        `send_embed` - This sends an embed.

        `**Parameters**``
        - `message` - The message you want to send.
        - `title` - The title of the embed.
        - `content` - The content of the embed.
        - `image` - The image you want to send. Recommended size: 807x216
        - `link` - The link you want to send. [Optional]

        `**Returns**`` - CMessage object.

        `**Example**``
        ```py
        @bot.command("embed")
        def embed(ctx: Context):
            return ctx.send_embed(
                message = "[c]",
                title = "Hello World!",
                content = "This is an embed.",
                image = "https://i.imgur.com/8ZQZ9Zm.png",
                link = "https://www.google.com"
            )
        ```
        """
        message: CMessage = self.__send_message__(
            content=message,
            attachedObject = {
                "title": title,
                "content": content,
                "mediaList": [[100, self.__handle_media__(media=image, media_value=True), None]],
                "link": link
                },
            extensions = {
                "mentionedArray": [{"uid": user} for user in mentioned] if mentioned else None
            })
        
        return message


    def __handle_media__(self, media: str, content_type: str = "image/jpg", media_value: bool = False) -> str:
        response = None
        
        try:
            if media.startswith("http"):
                response = get(media)
                response.raise_for_status()
                media = response.content
            else:
                media = open(media, "rb").read()
        except Exception as e:
            raise InvalidImage from e
        
        if content_type == "audio/aac":
            return self.encode_media(media)

        if media_value:
            return self.upload_media(media=media, content_type=content_type)

        if response and not response.headers.get("content-type").startswith("image"):
            raise InvalidImage

        return media
    

    def encode_media(self, file: bytes) -> str:
        return b64encode(file).decode()


    def upload_media(self, media: Union[str, BinaryIO], content_type: str = "image/jpg") -> str:
        return ApiResponse(self.request.handler(
            method = "POST",
            url = "/g/s/media/upload",
            data = media,
            content_type = content_type
            )).mediaValue


    @_run
    @__typing__
    def send_sticker(self, sticker_id: str) -> CMessage:
        """
        `send_sticker` - This sends a sticker.

        `**Parameters**``
        - `sticker_id` - The sticker ID you want to send.

        `**Returns**`` - CMessage object.

        `**Example**``
        ```py
        @bot.on_text_message()
        def on_text_message(ctx: Context):
            ctx.send_sticker(sticker_id="sticker_id")
        ```
        """
        sticker_id = sticker_id.replace("ndcsticker://", "") if sticker_id.startswith("ndcsticker://") else sticker_id
        message: CMessage = self.__send_message__(
            type=3,
            stickerId=sticker_id,
            mediaValue=f"ndcsticker://{sticker_id}"
            )
        
        return message


    @_run
    def send_image(self, image: str) -> CMessage:
        """
        `send_image` - This sends an image.

        `**Parameters**``
        - `image` - The image link or file you want to send.

        `**Returns**`` - CMessage object.

        `**Example**``
        ```py
        @bot.on_text_message()
        def on_text_message(ctx: Context):
            ctx.send_image(image="https://i.imgur.com/image.jpg")
        ```
        """
        message: CMessage = self.__send_message__(
            mediaType=100,
            mediaUploadValue=self.encode_media(
                self.__handle_media__(
                    media=image,
                    content_type="image/jpg",
                    media_value=False
            )))

        return message


    @_run
    def send_gif(self, gif: str) -> CMessage:
        """
        `send_gif` - This sends a gif.

        `**Parameters**``
        - `gif` - The gif link or file you want to send.

        `**Returns**`` - CMessage object.

        `**Example**``
        ```py
        @bot.on_text_message()
        def on_text_message(ctx: Context):
            ctx.send_gif(gif="https://i.imgur.com/image.gif")
        ```
        """
        message: CMessage = self.__send_message__(
            mediaType=100,
            mediaUploadValueContentType="image/gif",
            mediaUploadValue=self.encode_media(
                self.__handle_media__(
                    media=gif,
                    content_type="image/gif",
                    media_value=False
            )))
        
        return message


    @_run
    def send_audio(self, audio: str) -> CMessage:
        """
        `send_audio` - This sends an audio file.
        
        `**Parameters**``
        - `audio` - The audio link or file you want to send.
        
        `**Returns**`` - CMessage object.
        
        `**Example**``
        ```py
        @bot.on_text_message()
        def on_text_message(ctx: Context):
            ctx.send_audio(audio="output.mp3")
        ```
        """
        message: CMessage = self.__send_message__(
            type=2,
            mediaType=110,
            mediaUploadValue=self.__handle_media__(
                    media=audio,
                    content_type="audio/aac",
                    media_value=False
            ))
        
        return message


    @_run
    def join_chat(self, chatId: str=None) -> ApiResponse:
        """
        `join_chat` - This joins a chat.

        `**Example**``
        ```py
        @bot.on_text_message()
        def on_text_message(ctx: Context):
            ctx.join_chat(chatId="0000-0000-0000-0000")
        ```
        """
        return ApiResponse(self.request.handler(
            method="POST",
            url=f"/{self.communityId}/s/chat/thread/{chatId or self.chatId}/member/{self.userId}"
            ))


    @_run
    def leave_chat(self, chatId: str=None) -> ApiResponse:
        """
        `leave_chat` - This leaves a chat.

        `**Example**``
        ```py
        @bot.on_text_message()
        def on_text_message(ctx: Context):
            ctx.leave_chat(chatId="0000-0000-0000-0000")
        ```
        """
        return ApiResponse(self.request.handler(
            method="DELETE",
            url=f"/{self.communityId}/s/chat/thread/{chatId or self.chatId}/member/{self.userId}"
            ))


class EventHandler:
    """
    `EventHandler` - AKA where all the events are handled.

    `**Parameters**``
    - `session` - The session we are using.

    """
    def __init__(self):
        self.command_prefix:    str = self.command_prefix
        self._events:           dict = {}
        self._commands:         Commands = Commands()
        self.context:           Context = Context


    def register_event(self, event_name: str) -> Callable:
        def decorator(event_handler: Callable) -> Callable:
            self._events[event_name] = event_handler
            return event_handler
        return decorator


    def start_task(self, func):
        """`start_task` - This starts a task."""
        Thread(target=func).start()


    def _handle_task(self, func, interval):
        """
        `_handle_task` - This handles the task.
        
        `**Parameters**``
        - `func` - The function.
        - `interval` - The interval in seconds.
        
        `**Returns**`` - None
        """
        while True:
            if len(inspect_signature(func).parameters) == 0:
                func()
            else:
                func(self.community)
            delay(interval)


    def task(self, interval: int = 10):
        """
        `task` - This creates a task.

        `**Parameters**``
        - `interval` - The interval in seconds.

        `**Example**``
        ```py
        # This will print "Hello World!" every 10 seconds.
        @bot.task(interval=10)
        def task():
            print("Hello World!")

        # This will send a message to a chat every 120 seconds.
        @bot.task(interval=120)
        def task(community: Community):
            community.send_message(chatId=0000-0000-0000-0000, content="Hello World!")
        ```
        """

        def decorator(func):
            def wrapper():
                self._handle_task(func, interval)
            self.start_task(wrapper)
        return decorator


    def _set_parameters(self, context: Context, func: Callable, message: str = None) -> list:
        try:
            message = message if isinstance(message, str) else context.message.content
        except AttributeError:
            message = None

        try:
            username = context.author.username
        except AttributeError:
            username = None

        try:
            userId = context.author.userId
        except AttributeError:
            userId = None

        potential_parameters = {
            "ctx": context,
            "member": Member(context.author.json()),
            "message": message,
            "username": username,
            "userId": userId
        }

        return (
            potential_parameters.get(parameter)
            for parameter in inspect_signature(func).parameters
        )


    def emit(self, name: str, *args) -> None:
        """`emit` is a function that emits an event."""
        self._events[name](*args) if name in self._events else None


    def command(
        self,
        name: str=None,
        description: str=None,
        usage: str=None,
        aliases: list=[],
        cooldown: int=0,
        **kwargs
    ) -> Callable:
        """
        `command` - This creates a command.
        
        `**Command Parameters**``
        - `command_name` - The name of the command.
        - `command_description` - The description of the command.
        - `aliases` - The other names the command can be called by.
        - `cooldown` - The cooldown of the command in seconds.

        `**Function Parameters**``
        - `ctx` - The context of the command.
        - `member` - The Member(member) who called the command.
        - `message` - The message that called the command.
        - `username` - The username of the person who called the command.
        - `userId` - The userId of the person who called the command.

        Do I need a `command_description`?
            - No, you don't need a command description however it is recommended.
            - If you don't supply a command description the command will not show up in the help command.

        What are `aliases`?
            - The command can be called by the command name, aliases, or both.

        Is `cooldown` required?
            - No, you don't need a cooldown however it is recommended to avoid spam.

        What is the difference between `message` and `ctx.message.content`?
            - `ctx.message.content` contains the entire message.        
            - `message` contains the message without the command prefix.

        Do I need to supply all the parameters?
            - No, you only need to supply the parameters you want to use however `ctx` is required.
        
        `**Example**``
        ```py
        @bot.command(command_name="ping") # Command parameters.
        def ping(ctx: Context, message: str, username: str, userId: str): # Function parameters.
            print(f"{username}({userId}): {message}") # OUTPUT: "JohnDoe(0000-0000-0000-0000): !ping"
            return ctx.send(content="Pong!")

        @bot.command(command_name="ping", aliases=["alive", "test"]) # Command parameters.
        def ping(ctx: Context): # Function parameters.
            # This command can be called by "ping", "alive", and "test".
            return ctx.send(content="Pong!")

        @bot.command(command_name="ping", cooldown=5) # Command parameters.
        def ping(ctx: Context): # Function parameters.
            # This command can only be called every 5 seconds.
            return ctx.send(content="Pong!")

        @bot.command(command_name="say", command_description="This is a command that says something.") # Command parameters.
        def say(ctx: Context, message: str, username: str, userId: str): # Function parameters.
            bot.community.delete_message(chatId=ctx.chatId, messageId=ctx.message.chatId, comId=ctx.comId)
            return ctx.send(content=message)
        ```
        """
        
        if "command_name" in kwargs:
            self._is_deprecated("command_name", "name")
            name = kwargs["command_name"]

        elif name is None:
            raise ValueError("Please supply a name for the command. Example: @bot.command(name='ping')")

        if "command_description" in kwargs:
            self._is_deprecated("command_description", "description")
            description = kwargs["command_description"]

        def decorator(func: Callable) -> Callable:
            self._commands.add_command(
                Command(
                    func=func,
                    name=name,
                    description=description,
                    usage=usage,
                    aliases=aliases,
                    cooldown=cooldown
                ))
            return func
        return decorator


    def _is_deprecated(self, parameter: str, new_parameter: str):
        print(f"{Style.BRIGHT}{Fore.RED}WARNING:{Style.RESET_ALL} '{parameter}' is deprecated. Please use '{new_parameter}' instead.")


    def command_exists(self, command_name: str) -> bool:
        return any([
            command_name in self._commands.__command_names__(),
            command_name in self._commands.__command_aliases__()
            ])


    def fetch_command(self, command_name: str) -> Command:
        return self._commands.fetch_command(command_name)

    def _handle_command(self, data: Message, context: Context):
        """
        Handles commands.

        Args:
            self: The instance of the class.
            data (Message): The message data containing the command.
            context (Context): The context of the command.

        Returns:
            None or the response from the command function.

        Raises:
            None

        Examples:
            This function is internally called and does not have direct usage examples.
        """
        command_name = next(iter(data.content[len(self.command_prefix):].split(" ")))

        message = data.content[len(self.command_prefix) + len(command_name) + 1:]
        command = self._commands.fetch_command(command_name)
    
        if command is None:
            if command_name == "help" and data.content == f"{self.command_prefix}help":
                return context.reply(content=self._commands.__help__())
            
            if self._events.get("text_message"):
                return self._handle_all_events(event="text_message", data=data, context=context)

        if data.content[:len(self.command_prefix)] != self.command_prefix:
            return None

        try:
            response = self._check_cooldown(command.name, data, context)
        except AttributeError:
            return None

        if response != 403:
            func = command.func
            return func(*self._set_parameters(context=context, func=func, message=message))

        return None


    def _check_cooldown(self, command_name: str, data: Message, context: Context) -> None:
        """`_check_cooldown` is a function that checks if a command is on cooldown."""
        if self._commands.fetch_command(command_name).cooldown > 0:
            if self._commands.fetch_cooldown(command_name, data.author.userId) > time():
                context.reply(content=self._cooldown_message or f"You are on cooldown for {int(self._commands.fetch_cooldown(command_name, data.author.userId) - time())} seconds.")
                return 403
            
            self._commands.set_cooldown(
                command_name=command_name,
                cooldown=self._commands.fetch_command(command_name).cooldown,
                userId=data.author.userId
                )

        return 200

    def _add_cache(self, chatId: str, userId: str, content: str):
        with self.cache as cache:
            if cache.get(f"{chatId}_{userId}") is not None:
                self.cache.pop(f"{chatId}_{userId}")

            cache.add(
                key=f"{chatId}_{userId}",
                value=content,
                expire=90
                )


    def on_error(self):
        """
        `on_error` - This is an event that is called when an error occurs.
        """
        return self.register_event("error")


    def on_ready(self):
        """
        `on_ready` - This is an event that is called when the bot is ready to start handling events.
        """
        return self.register_event("ready")


    def on_text_message(self):
        """
        `on_text_message` - This is an event that is called when a text message is received in the chat.
        """
        return self.register_event("text_message")


    def _console_on_text_message(self):
        """
        `_console_on_text_message` - This is an event that is called when a text message is received in the console.
        """
        return self.register_event("_console_text_message")


    def on_image_message(self):
        """
        `on_image_message` - This is an event that is called when an image message is received in the chat.
        """
        return self.register_event("image_message")
        

    def on_youtube_message(self):
        """
        `on_youtube_message` - This is an event that is called when a YouTube message is received in the chat.
        """
        return self.register_event("youtube_message")


    def on_strike_message(self):
        """
        `on_strike_message` - This is an event that is called when a strike message is received in the chat.
        """
        return self.register_event("strike_message")


    def on_voice_message(self):
        """
        `on_voice_message` - This is an event that is called when a voice message is received in the chat.
        """
        return self.register_event("voice_message")


    def on_sticker_message(self):
        """
        `on_sticker_message` - This is an event that is called when a sticker message is received in the chat.
        """
        return self.register_event("sticker_message")


    def on_vc_not_answered(self):
        """
        `on_vc_not_answered` - This is an event that is called when a voice chat request is not answered.
        """
        return self.register_event("vc_not_answered")


    def on_vc_not_cancelled(self):
        """
        `on_vc_not_cancelled` - This is an event that is called when a voice chat request is not cancelled.
        """
        return self.register_event("vc_not_cancelled")



    def on_vc_not_declined(self):
        """
        `on_vc_not_declined` - This is an event that is called when a voice chat request is not declined.
        """
        return self.register_event("vc_not_declined")


    def on_video_chat_not_answered(self):
        """
        `on_video_chat_not_answered` - This is an event that is called when a video chat request is not answered.
        """
        return self.register_event("video_chat_not_answered")


    def on_video_chat_not_cancelled(self):
        """
        `on_video_chat_not_cancelled` - This is an event that is called when a video chat request is not cancelled.
        """
        return self.register_event("video_chat_not_cancelled")


    def on_video_chat_not_declined(self):
        """
        `on_video_chat_not_declined` - This is an event that is called when a video chat request is not declined.
        """
        return self.register_event("video_chat_not_declined")


    def on_avatar_chat_not_answered(self):
        """
        `on_avatar_chat_not_answered` - This is an event that is called when an avatar chat request is not answered.
        """
        return self.register_event("avatar_chat_not_answered")


    def on_avatar_chat_not_cancelled(self):
        """
        `on_avatar_chat_not_cancelled` - This is an event that is called when an avatar chat request is not cancelled.
        """
        return self.register_event("avatar_chat_not_cancelled")


    def on_avatar_chat_not_declined(self):
        """
        `on_avatar_chat_not_declined` - This is an event that is called when an avatar chat request is not declined.
        """
        return self.register_event("avatar_chat_not_declined")


    def on_delete_message(self):
        """
        `on_delete_message` - This is an event that is called when a message is deleted in the chat.
        """
        return self.register_event("delete_message")


    def on_member_join(self):
        """
        `on_member_join` - This is an event that is called when a member joins the chat.
        """
        return self.register_event("member_join")


    def on_member_leave(self):
        """
        `on_member_leave` - This is an event that is called when a member leaves the chat.
        """
        return self.register_event("member_leave")


    def on_chat_invite(self):
        """
        `on_chat_invite` - This is an event that is called when an invite is sent to the chat.
        """
        return self.register_event("chat_invite")


    def on_chat_background_changed(self):
        """
        `on_chat_background_changed` - This is an event that is called when the chat background is changed.
        """
        return self.register_event("chat_background_changed")


    def on_chat_title_changed(self):
        """
        `on_chat_title_changed` - This is an event that is called when the chat title is changed.
        """
        return self.register_event("chat_title_changed")


    def on_chat_icon_changed(self):
        """
        `on_chat_icon_changed` - This is an event that is called when the chat icon is changed.
        """
        return self.register_event("chat_icon_changed")


    def on_vc_start(self):
        """
        `on_vc_start` - This is an event that is called when a voice chat starts.
        """
        return self.register_event("vc_start")


    def on_video_chat_start(self):
        """
        `on_video_chat_start` - This is an event that is called when a video chat starts.
        """
        return self.register_event("video_chat_start")


    def on_avatar_chat_start(self):
        """
        `on_avatar_chat_start` - This is an event that is called when an avatar chat starts.
        """
        return self.register_event("avatar_chat_start")


    def on_vc_end(self):
        """
        `on_vc_end` - This is an event that is called when a voice chat ends.
        """
        return self.register_event("vc_end")


    def on_video_chat_end(self):
        """
        `on_video_chat_end` - This is an event that is called when a video chat ends.
        """
        return self.register_event("video_chat_end")


    def on_avatar_chat_end(self):
        """
        `on_avatar_chat_end` - This is an event that is called when an avatar chat ends.
        """
        return self.register_event("avatar_chat_end")


    def on_chat_content_changed(self):
        """
        `on_chat_content_changed` - This is an event that is called when the chat content is changed.
        """
        return self.register_event("chat_content_changed")


    def on_screen_room_start(self):
        """
        `on_screen_room_start` - This is an event that is called when a screen room starts.
        """
        return self.register_event("screen_room_start")


    def on_screen_room_end(self):
        """
        `on_screen_room_end` - This is an event that is called when a screen room ends.
        """
        return self.register_event("screen_room_end")


    def on_chat_host_transfered(self):
        """
        `on_chat_host_transfered` - This is an event that is called when the chat host is transferred.
        """
        return self.register_event("chat_host_transfered")


    def on_text_message_force_removed(self):
        """
        `on_text_message_force_removed` - This is an event that is called when a text message is forcefully removed.
        """
        return self.register_event("text_message_force_removed")


    def on_chat_removed_message(self):
        """
        `on_chat_removed_message` - This is an event that is called when a chat message is removed.
        """
        return self.register_event("chat_removed_message")


    def on_mod_deleted_message(self):
        """
        `on_mod_deleted_message` - This is an event that is called when a moderator deletes a message.
        """
        return self.register_event("mod_deleted_message")


    def on_chat_tip(self):
        """
        `on_chat_tip` - This is an event that is called when a tip is received in the chat.
        """
        return self.register_event("chat_tip")


    def on_chat_pin_announcement(self):
        """
        `on_chat_pin_announcement` - This is an event that is called when an announcement is pinned in the chat.
        """
        return self.register_event("chat_pin_announcement")


    def on_vc_permission_open_to_everyone(self):
        """
        `on_vc_permission_open_to_everyone` - This is an event that is called when voice chat permissions are set to open to everyone.
        """
        return self.register_event("vc_permission_open_to_everyone")


    def on_vc_permission_invited_and_requested(self):
        """
        `on_vc_permission_invited_and_requested` - This is an event that is called when voice chat permissions are set to invited and requested.
        """
        return self.register_event("vc_permission_invited_and_requested")


    def on_vc_permission_invite_only(self):
        """
        `on_vc_permission_invite_only` - This is an event that is called when voice chat permissions are set to invite only.
        """
        return self.register_event("vc_permission_invite_only")


    def on_chat_view_only_enabled(self):
        """
        `on_chat_view_only_enabled` - This is an event that is called when chat view only mode is enabled.
        """
        return self.register_event("chat_view_only_enabled")


    def on_chat_view_only_disabled(self):
        """
        `on_chat_view_only_disabled` - This is an event that is called when chat view only mode is disabled.
        """
        return self.register_event("chat_view_only_disabled")


    def on_chat_unpin_announcement(self):
        """
        `on_chat_unpin_announcement` - This is an event that is called when an announcement is unpinned in the chat.
        """
        return self.register_event("chat_unpin_announcement")


    def on_chat_tipping_enabled(self):
        """
        `on_chat_tipping_enabled` - This is an event that is called when chat tipping is enabled.
        """
        return self.register_event("chat_tipping_enabled")


    def on_chat_tipping_disabled(self):
        """
        `on_chat_tipping_disabled` - This is an event that is called when chat tipping is disabled.
        """
        return self.register_event("chat_tipping_disabled")


    def on_timestamp_message(self):
        """
        `on_timestamp_message` - This is an event that is called when a timestamp message is received in the chat.
        """
        return self.register_event("timestamp_message")


    def on_welcome_message(self):
        """
        `on_welcome_message` - This is an event that is called when a welcome message is received in the chat.
        """
        return self.register_event("welcome_message")


    def on_share_exurl_message(self):
        """
        `on_share_exurl_message` - This is an event that is called when a shared external URL message is received in the chat.
        """
        return self.register_event("share_exurl_message")


    def on_invite_message(self):
        """
        `on_invite_message` - This is an event that is called when an invite message is received in the chat.
        """
        return self.register_event("invite_message")


    def on_user_online(self):
        """
        `on_user_online` - This is an event that is called when a user comes online.
        """
        return self.register_event("user_online")


    def on_member_set_you_host(self):
        """
        `on_member_set_you_host` - This is an event that is called when you are set as the host of the chat.

        **Example:**
        ```python
        from pymino.ext import *
        chatId = "0000-0000-0000-0000"

        @bot.on_member_set_you_host()
        def member_set_you_host(notification: Notification):
            if notification.chatId == chatId:
                print("You are now the host")
                bot.community.send_message(chatId=chatId, content="I am now the host", comId=notification.comId)
        ```
        """
        return self.register_event("member_set_you_host")


    def on_member_set_you_cohost(self):
        """
        `on_member_set_you_cohost` - This is an event that is called when you are set as a cohost of the chat.
        
        **Example:**
        ```python
        from pymino.ext import *
        chatId = "0000-0000-0000-0000"
        
        @bot.on_member_set_you_cohost()
        def member_set_you_cohost(notification: Notification):
            if notification.chatId == chatId:
                print("You are now a cohost")
                bot.community.send_message(chatId=chatId, content="I am now a cohost", comId=notification.comId)
        ```
        """
        return self.register_event("member_set_you_cohost")


    def on_member_remove_your_cohost(self):
        """
        `on_member_remove_your_cohost` - This is an event that is called when you are removed as a cohost of the chat.
        
        **Example:**
        ```python
        from pymino.ext import *
        chatId = "0000-0000-0000-0000"
        
        @bot.on_member_remove_your_cohost()
        def member_remove_your_cohost(notification: Notification):
            if notification.chatId == chatId:
                print("You are no longer a cohost")
                bot.community.send_message(chatId=chatId, content="I am no longer a cohost", comId=notification.comId)
        ```
        """
        return self.register_event("member_remove_your_cohost")


    def _handle_all_events(self, event: str, data: Message, context: Context) -> None:
        func = self._events[event]
        return func(*self._set_parameters(context, func, data))


    def _handle_event(
        self,
        event: str,
        data: Union[Message, OnlineMembers, Notification, Context]
        ) -> Union[Context, None]:
        """
        `_handle_event` is a function that handles events.
        """
        with suppress(KeyError):

            if event == "text_message":
                context = self.context(data, self)
                if all([self.intents, not self.command_exists(
                    command_name=data.content[len(self.command_prefix):].split(" ")[0]
                    )]):
                        self._add_cache(data.chatId, data.author.userId, data.content)

                return self._handle_command(data=data, context=context)
            
            if event in self._events:
                context = self.context(data, self)

                if event in {
                    "user_online",
                    "member_set_you_host",
                    "member_set_you_cohost",
                    "member_remove_your_cohost",
                }:
                    return self._events[event](data)

                else:
                    return self._handle_all_events(event=event, data=data, context=context)
                
            return None